import platform
from pathlib import Path
from typing import Callable, List, Optional, Union

from PySide6.QtCore import QAbstractListModel, QModelIndex, QObject, QPersistentModelIndex, Qt
from PySide6.QtGui import QIcon
from typing_extensions import override

import conan_explorer.app as app  # using global module pattern
from conan_explorer import (
    INVALID_CONAN_REF,
    INVALID_PATH,
    USE_CONAN_WORKER_FOR_LOCAL_PKG_PATH_AND_INSTALL,
    USE_LOCAL_CACHE_FOR_LOCAL_PKG_PATH,
)
from conan_explorer.app.logger import Logger
from conan_explorer.conan_wrapper.conan_worker import ConanWorkerElement
from conan_explorer.conan_wrapper.types import ConanRef
from conan_explorer.settings import AUTO_INSTALL_QUICKLAUNCH_REFS
from conan_explorer.ui.common import (
    extract_icon,
    get_asset_image_path,
    get_icon_from_image_file,
    get_themed_asset_icon,
)

from .config import UiAppGridConfig, UiAppLinkConfig, UiTabConfig


class UiAppGridModel(UiAppGridConfig, QObject):
    def __init__(self, *args, **kwargs):
        UiAppGridConfig.__init__(self, *args, **kwargs)
        QObject.__init__(self)
        self.tabs: List[UiTabModel]

    def save(self):
        if self.parent:
            self.parent.save()

    def load(self, ui_config: UiAppGridConfig, parent) -> "UiAppGridModel":
        super().__init__(ui_config.tabs)
        self.parent = parent
        # load all submodels
        tabs_model = []
        for tab_config in self.tabs:
            tabs_model.append(UiTabModel().load(tab_config, self))
        self.tabs = tabs_model
        return self

    def get_all_conan_worker_elements(self) -> List[ConanWorkerElement]:
        """Helper function to generate a unique list of all ConanWorkerElements"""
        conan_refs: List[ConanWorkerElement] = []
        for tab in self.tabs:
            for app_model in tab.apps:
                conan_worker_element: ConanWorkerElement = {
                    "ref_pkg_id": app_model.conan_ref,
                    "settings": {},
                    "options": app_model.conan_options,
                    "update": False,
                    "profile": "",
                    "auto_install": True,
                }
                if conan_worker_element not in conan_refs:
                    conan_refs.append(conan_worker_element)
        return conan_refs


class UiTabModel(UiTabConfig, QAbstractListModel):
    """Representation of a tab entry of the config schema"""

    def __init__(self, *args, **kwargs):
        """Create an empty AppModel on init, so we can load it later"""
        UiTabConfig.__init__(self, *args, **kwargs)
        QAbstractListModel.__init__(self)
        self.parent: UiAppGridModel
        self.apps: List[UiAppLinkModel]

    def load(self, config: UiTabConfig, parent: "UiAppGridModel") -> "UiTabModel":
        super().__init__(config.name, config.apps)
        self.parent = parent
        # load all submodels
        apps_model = []
        for app_config in self.apps:
            apps_model.append(UiAppLinkModel().load(app_config, self, trigger_update=False))
        self.apps = apps_model
        return self

    def save(self):
        if self.parent:  # delegate to top
            self.parent.save()

    @override
    def data(self, index, role):
        if role == Qt.ItemDataRole.DisplayRole:
            return self.apps[index.row()].name
        if role == Qt.ItemDataRole.UserRole:
            return self.apps[index.row()]

    @override
    def setData(self, index, value, role):
        if role == Qt.ItemDataRole.UserRole:
            self.apps[index.row()] = value

    @override
    def rowCount(self, parent=None) -> int:
        return len(self.apps)

    @override
    def columnCount(self, parent: "QModelIndex | QPersistentModelIndex") -> int:
        return 1

    @override
    def insertRow(self, row: int, parent=QModelIndex()) -> bool:
        self.apps.insert(row, UiAppLinkModel())
        return super().insertRow(row, parent=parent)

    @override
    def removeRow(self, row: int, parent=QModelIndex()) -> bool:
        self.apps.pop(row)
        return super().removeRow(row, parent=parent)

    @override
    def moveRow(
        self,
        source_parent: Union[QModelIndex, QPersistentModelIndex],
        source_row: int,
        destination_parent: Union[QModelIndex, QPersistentModelIndex],
        destination_child: int,
    ) -> bool:
        app_to_move = self.apps[source_row]
        self.apps.insert(destination_child, app_to_move)
        if source_row < destination_child:
            self.apps.pop(source_row)
        else:
            self.apps.pop(source_row + 1)
        return super().moveRow(source_parent, source_row, destination_parent, destination_child)


class UiAppLinkModel(UiAppLinkConfig):
    """Representation of an app link entry of the config"""

    INVALID_DESCR = "NA"
    OFFICIAL_RELEASE = "<official release>"  # to be used for "_" channel
    OFFICIAL_USER = "<official user>"  # to be used for "_" user

    def __init__(self, *args, **kwargs):
        """Create an empty AppModel on init, so we can load it later"""
        # internal repr for vars which have other types or need to be manipulated
        self.conan_options = {}
        self.package_folder = Path(INVALID_PATH)
        self._executable_path = Path(INVALID_PATH)
        self._conan_ref = INVALID_CONAN_REF
        self._conan_file_reference = ConanRef.loads(self._conan_ref)
        self._available_refs: List[ConanRef] = []
        self._executable = ""
        self._icon: str = ""
        self.lock_changes = (
            True  # values do not emit events. used when multiple changes are needed
        )
        self.parent = UiTabModel("default")
        # can be registered from external function to notify when conan infos habe been fetched asynchronnaly
        self._update_cbk_func: Optional[Callable] = None
        # calls public functions, every internal variable needs to initialitzed
        super().__init__(*args, **kwargs)  # empty init

    def load(
        self, config: UiAppLinkConfig, parent: UiTabModel, trigger_update=True
    ) -> "UiAppLinkModel":
        self.parent = parent
        if trigger_update:
            self.lock_changes = False  # don't trigger updates to worker - will be done in batch

        super().__init__(
            config.name,
            config.executable,
            config.icon,
            config.is_console_application,
            config.args,
            config.conan_options,
            config.conan_ref,
        )
        self.lock_changes = False  # unlock for futher use now that everything is loaded
        return self

    def save(self):
        if self.parent:  # delegate to top
            self.parent.save()

    def load_from_cache(self):
        if not app.conan_api:
            return
        # get all info from cache
        self.set_available_packages(
            app.conan_api.info_cache.get_similar_pkg_refs(
                self._conan_file_reference.name, user="*"
            )
        )
        pkg_path = Path(INVALID_PATH)
        if USE_LOCAL_CACHE_FOR_LOCAL_PKG_PATH:
            pkg_path = app.conan_api.info_cache.get_local_package_path(
                self._conan_file_reference
            )
            if not pkg_path.exists():
                _, pkg_path = app.conan_api.get_best_matching_local_package_path(
                    self._conan_file_reference, self.conan_options
                )
            if self.conan_options:
                pkg_info = app.conan_api.get_local_pkg_from_path(
                    self._conan_file_reference, pkg_path
                )
                # user options should be a subset of full pkg options
                if pkg_info:
                    if not self.conan_options.items() <= pkg_info.get("options", {}).items():
                        return
        if (
            not pkg_path.exists() and not USE_CONAN_WORKER_FOR_LOCAL_PKG_PATH_AND_INSTALL
        ):  # last chance to get path
            _, pkg_path = app.conan_api.get_path_or_auto_install(
                self._conan_file_reference, self.conan_options
            )

        self.set_package_folder(pkg_path, quiet=True)

    def register_update_callback(self, update_func: Callable):
        """This callback can be used to update the gui after new conan info was received"""
        self._update_cbk_func = update_func

    @property
    def conan_file_reference(self) -> ConanRef:
        return self._conan_file_reference

    @property
    def conan_ref(self) -> str:
        """The conan reference to be used for the link."""
        return self._conan_ref

    @conan_ref.setter
    def conan_ref(self, new_value: str):
        try:
            self._conan_file_reference = ConanRef.loads(new_value)
        except Exception:  # invalid ref
            Logger().debug("Invalid ref: %s", new_value)
            return
        # add conan ref to worker
        if (
            self._conan_ref != new_value
            and new_value != INVALID_CONAN_REF
            and self._conan_file_reference.version != self.INVALID_DESCR
            and self._conan_file_reference.channel != self.INVALID_DESCR
        ):  # don't put it for init
            # invalidate old entries, which are dependent on the conan ref - only for none invalid refs
            self._conan_ref = new_value
            self.load_from_cache()
            if self.parent and self.parent.parent and not self.lock_changes:
                self.trigger_conan_update()
        self._conan_ref = new_value

    def trigger_conan_update(self):
        if not app.active_settings.get_bool(AUTO_INSTALL_QUICKLAUNCH_REFS):
            return
        try:
            conan_worker_element: ConanWorkerElement = {
                "ref_pkg_id": str(self._conan_ref),
                "settings": {},
                "profile": "",
                "options": self.conan_options,
                "update": True,
                "auto_install": True,
            }
            app.conan_worker.put_ref_in_install_queue(
                conan_worker_element, self.emit_conan_pkg_signal_callback
            )
        except Exception as e:
            # errors happen fairly often, keep going
            Logger().warning(f"Conan reference invalid {str(e)}")

    def emit_conan_pkg_signal_callback(self, conan_ref, pkg_id):
        if not self.parent.parent.parent.conan_pkg_installed:
            return
        self.parent.parent.parent.conan_pkg_installed.emit(conan_ref, pkg_id)

    @property
    def version(self) -> str:
        """Version, as specified in the conan ref"""
        return str(self._conan_file_reference.version)

    @version.setter
    def version(self, new_value: str):
        user = self._conan_file_reference.user
        channel = self._conan_file_reference.channel
        if not self._conan_file_reference.user or not self._conan_file_reference.channel:
            user = "_"
            channel = "_"  # both must be unset if channel is official
        self.conan_ref = str(
            ConanRef(self._conan_file_reference.name, new_value, user, channel)
        )

    @classmethod
    def convert_to_disp_channel(cls, channel: Optional[str]) -> str:
        """Substitute _ for official channel string"""
        if not channel or channel == "_":
            return cls.OFFICIAL_RELEASE
        return channel

    @property
    def user(self) -> str:
        """User, as specified in the conan ref"""
        return self._convert_to_disp_user(self._conan_file_reference.user)

    @user.setter
    def user(self, new_value: str):
        channel = self._conan_file_reference.channel
        if new_value == self.OFFICIAL_USER or not new_value:
            new_value = "_"
            channel = "_"  # both must be unset if channel is official
        if not channel or channel == "_":
            channel = "NA"
        self.conan_ref = str(
            ConanRef(
                self._conan_file_reference.name,
                self._conan_file_reference.version,
                new_value,
                channel,
            )
        )

    @classmethod
    def _convert_to_disp_user(cls, user: Optional[str]) -> str:
        """Substitute _ for official user string"""
        if not user or user == "_":
            return cls.OFFICIAL_USER
        return user

    @property
    def channel(self) -> str:
        """Channel, as specified in the conan ref"""
        return self.convert_to_disp_channel(self._conan_file_reference.channel)

    @channel.setter
    def channel(self, new_value: str):
        user = self.user
        # even when changing to another channel, it will reset, user or whole ref has to be changed
        if new_value == self.OFFICIAL_RELEASE or not new_value or user == self.OFFICIAL_USER:
            new_value = "_"
            user = "_"  # both must be unset if channel is official
        self.conan_ref = str(
            ConanRef(
                self._conan_file_reference.name,
                self._conan_file_reference.version,
                user,
                new_value,
            )
        )

    @property
    def versions(self) -> List[str]:
        """All versions (sorted) for the current name and user"""
        versions = set()
        for ref in self._available_refs:
            versions.add(ref.version)
        versions_list = list(versions)
        versions_list.sort()
        return versions_list

    @property
    def users(self) -> List[str]:
        """All users (sorted) for the current name and verion"""
        users = set()
        # users.add(self.user)
        for ref in self._available_refs:
            if ref.version == self.version:
                users.add(self._convert_to_disp_user(ref.user))
        users_list = list(users)
        users_list.sort()
        return users_list

    @property
    def channels(self) -> List[str]:
        """All channels (sorted) for the current version and user only"""
        channels = set()
        channels.add(self.channel)
        for ref in self._available_refs:
            if (
                ref.version == self.version
                and self._convert_to_disp_user(ref.user) == self.user
            ):
                channels.add(self.convert_to_disp_channel(ref.channel))
        # extra handling for only one existing channel found
        if len(channels) == 2 and self.INVALID_DESCR in channels:
            channels.remove(self.INVALID_DESCR)
        channels_list = list(channels)
        channels_list.sort()
        return channels_list

    @property
    def executable(self) -> str:
        """The executable for this link to trigger"""
        return self._executable

    @executable.setter
    def executable(self, new_value: str):
        self._executable = new_value

    def get_executable_path(self) -> Path:
        if not self._executable or not self.package_folder.exists():
            return Path(INVALID_PATH)
        path = Path(self._executable)
        full_path = self.resolve_executable_path(path)
        self._executable_path = full_path

        return self._executable_path

    def resolve_executable_path(self, exe_rel_path: Path) -> Path:
        # adjust path on windows, if no file extension is given
        if platform.system() == "Windows":
            possible_matches = self.package_folder.glob(str(exe_rel_path.as_posix()) + "*")
            match_found = False
            match = Path(INVALID_PATH)
            try:
                for match in possible_matches:
                    # don't allow for ambiguity!
                    if match_found:
                        Logger().error(
                            (
                                f"Multiple candidates found for {exe_rel_path}"
                                f"in {self.name}: e.g. {str(match.name)}"
                            )
                        )
                    match_found = True
                if not match_found:
                    Logger().debug(
                        "Can't find file in package %s:\n\t%s",
                        self.conan_ref,
                        str(exe_rel_path),
                    )
                else:
                    return match
            except NotImplementedError:
                Logger().error("Absolute path not allowed!")
            return Path(INVALID_PATH)
        else:
            possible_match = self.package_folder / exe_rel_path
            if not possible_match.exists():
                Logger().debug(
                    "Can't find file in package %s:\n\t%s", self.conan_ref, str(exe_rel_path)
                )
                return Path(INVALID_PATH)
            return possible_match

    @property
    def icon(self) -> str:
        """Internal representation of Icon to display on the link"""
        return self._icon

    @icon.setter
    def icon(self, new_value: str):
        self._icon = new_value

    def _eval_icon_path(self) -> Path:
        icon_path = Path(INVALID_PATH)
        # relative to package - migrate from old setting
        # config path will be deprecated
        if self._icon.startswith("//"):
            self._icon = self._icon.replace("//", "./")
        if self._icon and not Path(self._icon).is_absolute():
            icon_path = self.package_folder / self._icon
        else:  # absolute path
            icon_path = Path(self._icon)
        try:
            icon_path = icon_path.resolve()
        except Exception as e:
            Logger().debug("Can't resolve path of %s: %s", str(icon_path), str(e))
        if not icon_path.exists():
            if self.get_executable_path().exists():
                icon = "app.svg"
            else:
                icon = "no-access.svg"
            icon_path = Path(get_asset_image_path("icons/" + icon))
        return icon_path

    def get_icon(self) -> QIcon:
        """Get an icon based on icon path"""
        icon = None

        if not self._icon:
            icon = extract_icon(self.get_executable_path())
        else:
            icon = get_icon_from_image_file(self._eval_icon_path())

        # default icon, until package path is updated
        if icon.isNull():
            icon = get_themed_asset_icon("icons/app.svg")
            if self._icon:  # user input given -> warning
                Logger().debug("Can't find icon %s for '%s'", self._icon, self.name)
        return icon

    def set_package_folder(self, package_folder: Path, quiet=False):
        """
        Sets package path and all dependent paths.
        Use, when conan operation is done and paths can be validated.
        Usually to be called from conan worker.
        """

        if USE_LOCAL_CACHE_FOR_LOCAL_PKG_PATH:
            if self.package_folder != package_folder:
                app.conan_api.info_cache.update_local_package_path(
                    self._conan_file_reference, package_folder
                )
        self.package_folder = package_folder

        if not quiet and not package_folder.exists():
            Logger().info(
                f"Can't find a package for <b>{str(self.conan_ref)}"
                + f"</b> and options {repr(self.conan_options)} <b>locally</b>"
            )

        # call registered update callback
        if self._update_cbk_func:
            self._update_cbk_func()

    def set_available_packages(self, available_refs: List[ConanRef]):
        """
        Set all other available packages.
        Usually to be called from conan worker.
        """
        self._available_refs = available_refs

        # call registered update callback
        if self._update_cbk_func:
            self._update_cbk_func()



import pprint
from typing import List, Optional

from PySide6.QtCore import QModelIndex, QObject, Qt, SignalInstance
from PySide6.QtWidgets import QApplication, QListWidget, QPushButton, QTextBrowser, QTreeView

from conan_explorer.app import LoaderGui  # using global module pattern
from conan_explorer.ui.common import show_conanfile
from conan_explorer.ui.dialogs import ConanInstallDialog
from conan_explorer.ui.dialogs.pkg_diff.diff import PkgDiffDialog
from conan_explorer.ui.widgets.conan_line_edit import ConanRefLineEdit

from .model import PROFILE_TYPE, PkgSearchModel, SearchedPackageTreeItem


class ConanSearchController(QObject):
    def __init__(
        self,
        view: QTreeView,
        search_line: ConanRefLineEdit,
        search_button: QPushButton,
        remote_list: QListWidget,
        detail_view: QTextBrowser,
        conan_pkg_installed: Optional[SignalInstance],
        conan_pkg_removed: Optional[SignalInstance],
    ) -> None:
        super().__init__(view)
        self._view = view
        self._search_line = search_line
        self._search_button = search_button
        self._remote_list = remote_list
        self._detail_view = detail_view
        self._model = PkgSearchModel()
        self._loader = LoaderGui(self)
        self.conan_pkg_installed = conan_pkg_installed
        self.conan_pkg_removed = conan_pkg_removed

    def on_search(self):
        """Search for the user entered text by re-initing the model"""
        # IMPORTANT! if put in async loading, the pyqt signal of the model will be created in another Qt thread
        # and not be able to emit to the GUI thread.
        if not self._search_button.isEnabled():
            return
        self._model = PkgSearchModel(self.conan_pkg_installed, self.conan_pkg_removed)
        self._loader.load(
            self._view,
            self._load_search_model,
            (),
            self._finish_load_search_model,
            "Searching for packages...",
        )

        # reset info text
        self._detail_view.setText("")

    def _load_search_model(self):
        """Initialize tree view model by searching in conan"""
        self._model.setup_model_data(
            self._search_line.text(),
            self.get_selected_remotes(),
            self._loader.loading_string_signal,
        )

    def _finish_load_search_model(self, ret=None):
        """After conan search adjust the view"""
        self._view.setModel(self._model.proxy_model)
        self._resize_search_result_columns()
        self._view.sortByColumn(1, Qt.SortOrder.AscendingOrder)  # sort by remote at default
        self._view.selectionModel().selectionChanged.connect(self.on_package_selected)
        self._search_line.load_completion_refs()

    def on_package_selected(self):
        """Display package info only for pkg type"""
        items = self.get_selected_source_items()
        if len(items) != 1:
            return
        item = items[0]
        if item.type != PROFILE_TYPE:
            return
        pkg_info = pprint.pformat(item.pkg_info).translate(
            {ord("{"): None, ord("}"): None, ord("'"): None}
        )
        self._detail_view.setText(pkg_info)

    def on_copy_ref_requested(self):
        """Copy the selected reference to the clipboard"""
        combined_ref = self.get_selected_combined_ref()
        QApplication.clipboard().setText(combined_ref)

    def on_show_conanfile_requested(self):
        """Show the conanfile by downloading and opening with the associated program"""
        combined_ref = self.get_selected_combined_ref()
        conan_ref = combined_ref.split(":")[0]
        loader = LoaderGui(self)
        loader.load(
            self._view, show_conanfile, (conan_ref,), loading_text="Opening Conanfile..."
        )

    def on_install_pkg_requested(self):
        """Spawn the Conan install dialog"""
        combined_ref = self.get_selected_combined_ref()
        self.install(combined_ref)

    def on_install_button(self):
        self.install(self._search_line.text())

    def on_diff_requested(self):
        dialog = PkgDiffDialog(self._view)
        items = self.get_selected_source_items()
        if len(items) < 2:
            return
        for item in items:
            if not item.pkg_info:
                continue
            dialog.add_diff_item(item.pkg_info)
        dialog.show()

    def install(self, ref):
        dialog = ConanInstallDialog(self._view, ref, self.conan_pkg_installed)
        dialog.show()

    def get_selected_remotes(self) -> List[str]:
        """Returns the user selected remotes"""
        selected_remotes = []
        for i in range(self._remote_list.count()):
            item = self._remote_list.item(i)
            if item is None:
                continue
            if item.checkState() == Qt.CheckState.Checked:
                selected_remotes.append(item.text())
        return selected_remotes

    def get_selected_combined_ref(self) -> str:
        """Returns the user selected ref in <ref>:<id> format"""
        # no need to map from postition, since rightclick selects a single item
        source_items = self.get_selected_source_items()
        if len(source_items) != 1:
            return ""
        source_item = source_items[0]
        conan_ref_item = source_item
        id_str = ""
        if source_item.type == PROFILE_TYPE:
            conan_ref_item = source_item.parent()
            if source_item.pkg_info is not None:
                id_str = ":" + source_item.pkg_info.get("id", "")
        if not conan_ref_item:
            return ""
        return conan_ref_item.item_data[0] + id_str

    def get_selected_source_items(self) -> List[SearchedPackageTreeItem]:
        """Gets the selected item from a view"""
        indexes = self._get_selected_source_indexes()
        source_items: List[SearchedPackageTreeItem] = []
        for index in indexes:
            source_items.append(index.model().mapToSource(index).internalPointer())  # type: ignore

        return source_items

    def _get_selected_source_indexes(self) -> List[QModelIndex]:
        indexes = []
        for index in self._view.selectedIndexes():
            if index.column() == 0:  # we only need a row once
                indexes.append(index)
        return indexes

    def _resize_search_result_columns(self):
        for i in reversed(range(self._model.root_item.column_count() - 1)):
            self._view.resizeColumnToContents(i)



from PySide6.QtWidgets import QTextBrowser


class PlainTextPasteBrowser(QTextBrowser):
    def insertFromMimeData(self, source):
        """Paste text always as plain, not html"""
        self.insertPlainText(source.text())



import sys
import traceback
from types import TracebackType
from typing import Optional

from conan_explorer import user_save_path


def bug_dialog_exc_hook(
    exctype: "Optional[type[BaseException]]",
    excvalue: Optional[BaseException],
    tb: "Optional[TracebackType]",
):
    """App crash handling:
    print, write log and if the GUI still works show crash dialog"""

    print("Application crashed")
    error_text = f"ERROR: {str(exctype)} {excvalue}"
    with open(user_save_path / "crash.log", "w") as fd:
        fd.write(error_text + "\n")
        traceback.print_tb(tb, limit=10, file=fd)
    # print it too the console too
    print(error_text)
    traceback.print_tb(tb, limit=10)
    try:
        from conan_explorer.ui.dialogs import show_bug_reporting_dialog

        show_bug_reporting_dialog(excvalue, tb)
    except Exception as e:
        print(str(e))
        # gui does not work anymore - nothing to do
        sys.exit(2)
    sys.exit(1)



from __future__ import annotations
from contextlib import redirect_stderr, redirect_stdout
from dataclasses import dataclass
from pathlib import Path
import os
import platform
import pprint
from typing import Any, Dict, List, Literal, Optional, TypedDict, Union
from typing_extensions import TypeAlias

from conan_explorer import conan_version

from conans.errors import ConanException


if conan_version.major == 1:
    from conans.model.ref import ConanFileReference, PackageReference # type: ignore
    from conans.paths.package_layouts.package_editable_layout import PackageEditableLayout
    if platform.system() == "Windows":
        from conans.util.windows import CONAN_REAL_PATH, CONAN_LINK
else:
    from conans.model.recipe_ref import RecipeReference as ConanFileRef  # type: ignore
    from conans.model.package_ref import PkgReference  # type: ignore
    class PackageReference(PkgReference): # type: ignore
        """ Compatibility class for changed package_id attribute """
        ref: ConanRef
        
        @property
        def id(self):
            return self.package_id

        @staticmethod
        def loads(text: str) -> ConanPkgRef:
            pkg_ref = PkgReference.loads(text)
            return PackageReference(pkg_ref.ref, pkg_ref.package_id, 
                                    pkg_ref.revision,pkg_ref.timestamp)

    class ConanFileReference(ConanFileRef):
        name: str
        version: str
        user: Optional[str]
        channel: Optional[str]

        @staticmethod
        def loads(text: str, validate=True) -> ConanRef:
            # add back support for @_/_ canonical refs to handle this uniformly
            # Simply remove it before passing it to ConanFileRef
            if text.endswith("@_/_"):
                text = text.replace("@_/_", "")
            ref: ConanRef = ConanFileRef().loads(text) # type: ignore
            
            if validate:
                # validate_ref creates an own output stream which can't log to console
                # if it is running as a gui application
                devnull = open(os.devnull, 'w')
                with redirect_stdout(devnull):
                    with redirect_stderr(devnull):
                        ref.validate_ref(allow_uppercase=True)
            return ref

@dataclass
class Remote():
    name: str
    url: str
    verify_ssl: bool
    disabled: bool
    allowed_packages: Optional[List[str]] = None

from conans.errors import ConanException

ConanRef: TypeAlias = ConanFileReference
ConanPkgRef: TypeAlias = PackageReference
ConanOptions: TypeAlias = Dict[str, Any]
ConanAvailableOptions: TypeAlias = Dict[str, Union[List[Any], Literal["ANY"]]]
ConanSettings: TypeAlias = Dict[str, str]
ConanPackageId: TypeAlias = str
ConanPackagePath: TypeAlias = Path

    
class ConanPkg(TypedDict, total=False):
    """ Dummy class to type conan returned package dicts """

    id: ConanPackageId
    options: ConanOptions
    settings: ConanSettings
    requires: List[Any]
    outdated: bool


@dataclass
class EditablePkg():
    conan_ref: str
    path: str # path to conanfile or folder
    output_folder: Optional[str]

def pretty_print_pkg_info(pkg_info: ConanPkg) -> str:
    return pprint.pformat(pkg_info).translate(
        {ord("{"): None, ord("}"): None, ord("'"): None})


class LoggerWriter:
    """
    Dummy stream to log directly to a logger object, when writing in the stream.
    Used to redirect custom stream from Conan. 
    Adds a prefix to do some custom formatting in the Logger.
    """

    def __init__(self, level, prefix: str):
        self.level = level
        self._prefix = prefix

    def write(self, message: str):
        if message != '\n':
            self.level(self._prefix + message.strip("\n"))

    def flush(self):
        """ For interface compatiblity """
        pass


def create_key_value_pair_list(input_dict: Dict[str, Any]) -> List[str]:
    """
    Helper to create name=value string list from dict
    Filters "ANY" options.
    """
    res_list: List[str] = []
    if not input_dict:
        return res_list
    for name, value in input_dict.items():
        value = str(value)
        # this is not really safe, but there can be wild values...
        if "any" in value.lower() or "none" in value.lower():
            continue
        res_list.append(name + "=" + value)
    return res_list



"""Use constants in class, so they don't need to be separately accessed"""

from abc import abstractmethod
from typing import List, Optional

from conan_explorer import PathLike

GENERAL_SECTION_NAME = "General"
VIEW_SECTION_NAME = "View"
PLUGINS_SECTION_NAME = "Plugins"

# Constants for option names (value is the entry name/id)
### General
LAST_CONFIG_FILE = "last_config_file"
FILE_EDITOR_EXECUTABLE = "file_editor"
AUTO_INSTALL_QUICKLAUNCH_REFS = "auto_install_quicklaunch"
DEFAULT_INSTALL_PROFILE = "default_install_profile"

### View
FONT_SIZE = "font_size"
GUI_STYLE = "style"
# style choices
GUI_STYLE_MATERIAL = "material"
GUI_STYLE_FLUENT = "fluent"

GUI_MODE = "mode"
# mode choices
GUI_MODE_DARK = "dark"
GUI_MODE_LIGHT = "light"

WINDOW_SIZE = "window_size"
CONSOLE_SPLIT_SIZES = "console_split_sizes"
LAST_VIEW = "last_view"
AUTO_OPEN_LAST_VIEW = "auto_open_last_view"

# Implementation types for factory
SETTINGS_INI_TYPE = "ini"
SETTINGS_QT_TYPE = "qt"  # Try this out, see issue #56


def settings_factory(type: str, source: PathLike) -> "SettingsInterface":
    if type == SETTINGS_INI_TYPE:
        from conan_explorer.settings.ini_file import IniSettings

        implementation = IniSettings(source)
    elif type == SETTINGS_QT_TYPE:
        raise NotImplementedError
    else:
        raise NotImplementedError
    return implementation


# Interface for Settings to implement


class SettingsInterface:
    """
    Abstract Class to implement settings mechanisms.
    Source artefact location is passed by constructor and not changeable.
    """

    @abstractmethod
    def set_auto_save(self, value: bool):
        """Enable auto save feature, so every value change triggers a save."""
        raise NotImplementedError

    @abstractmethod
    def save(self):
        """Save all user modifiable settings."""
        raise NotImplementedError

    @abstractmethod
    def get_settings_from_node(self, node: str) -> List[str]:
        """Get all settings names from a hierachical node.
        If it is non-hierarchical, the name arg should be ignored and all settings returned.
        """
        raise NotImplementedError

    @abstractmethod
    def get(self, name: str) -> "str | int | float | bool":
        """Default getter, not good for typing"""
        raise NotImplementedError

    @abstractmethod
    def get_string(self, name: str) -> str:
        """Get a string cast to a string"""
        raise NotImplementedError

    @abstractmethod
    def get_int(self, name: str) -> int:
        """Get a string cast to an int"""
        raise NotImplementedError

    @abstractmethod
    def get_float(self, name: str) -> float:
        """Get a string cast to a float"""
        raise NotImplementedError

    @abstractmethod
    def get_bool(self, name: str) -> bool:
        """Get a string cast to a bool"""
        raise NotImplementedError

    @abstractmethod
    def set(self, name: str, value: "str|int|float|bool"):
        """Set the value of an existing setting"""
        raise NotImplementedError

    @abstractmethod
    def add(self, name: str, value: "str|int|float|bool", node: Optional[str] = None):
        """Add a new setting"""
        raise NotImplementedError

    @abstractmethod
    def remove(self, name: str):
        """Remove a setting"""
        raise NotImplementedError



from PySide6.QtWidgets import QMessageBox

from .clickable_icon import ClickableIcon
from .conan_line_edit import ConanRefLineEdit
from .text_broswer import PlainTextPasteBrowser
from .toggle import AnimatedToggle


class WideMessageBox(QMessageBox):
    """MessageBox with more width"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self._width = 600  # default

    def setWidth(self, width):
        self._width = width

    def resizeEvent(self, event):
        _result = super().resizeEvent(event)

        self.setFixedWidth(self._width)
        return _result



# base dialogs from which other dialogs can inherit
from .question_with_item_list import QuestionWithItemListDialog  # noqa
from .reorder import ReorderController, ReorderDialog, ReorderingModel  # noqa

# specific dialogs
from .file_editor_selection import FileEditorSelDialog
from .conan_cache_cleanup import ConanCacheCleanupDialog
from .conan_install import ConanInstallDialog
from .conan_remove import ConanRemoveDialog
from .crash import show_bug_reporting_dialog

__all__ = [
    "QuestionWithItemListDialog",
    "ReorderController",
    "ReorderDialog",
    "ReorderingModel",
    "show_bug_reporting_dialog",
    "ConanCacheCleanupDialog",
    "ConanInstallDialog",
    "ConanRemoveDialog",
    "FileEditorSelDialog",
]



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'editable_edit_dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QDialog, QDialogButtonBox,
    QGridLayout, QLabel, QLayout, QLineEdit,
    QPushButton, QSizePolicy, QVBoxLayout, QWidget)

from conan_explorer.ui.widgets.conan_line_edit import ConanRefLineEdit

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(645, 146)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(Dialog.sizePolicy().hasHeightForWidth())
        Dialog.setSizePolicy(sizePolicy)
        Dialog.setModal(True)
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.edit_grid = QGridLayout()
        self.edit_grid.setObjectName(u"edit_grid")
        self.edit_grid.setSizeConstraint(QLayout.SetMinAndMaxSize)
        self.output_folder_label = QLabel(Dialog)
        self.output_folder_label.setObjectName(u"output_folder_label")

        self.edit_grid.addWidget(self.output_folder_label, 2, 0, 1, 1)

        self.path_browse_button = QPushButton(Dialog)
        self.path_browse_button.setObjectName(u"path_browse_button")
        self.path_browse_button.setMaximumSize(QSize(50, 16777215))

        self.edit_grid.addWidget(self.path_browse_button, 1, 2, 1, 1)

        self.name_label = QLabel(Dialog)
        self.name_label.setObjectName(u"name_label")
        self.name_label.setWordWrap(True)

        self.edit_grid.addWidget(self.name_label, 0, 0, 1, 1)

        self.path_line_edit = QLineEdit(Dialog)
        self.path_line_edit.setObjectName(u"path_line_edit")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.path_line_edit.sizePolicy().hasHeightForWidth())
        self.path_line_edit.setSizePolicy(sizePolicy1)
        self.path_line_edit.setMinimumSize(QSize(0, 0))
        self.path_line_edit.setFrame(True)

        self.edit_grid.addWidget(self.path_line_edit, 1, 1, 1, 1)

        self.path_label = QLabel(Dialog)
        self.path_label.setObjectName(u"path_label")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.path_label.sizePolicy().hasHeightForWidth())
        self.path_label.setSizePolicy(sizePolicy2)
        self.path_label.setMinimumSize(QSize(0, 0))
        self.path_label.setWordWrap(True)

        self.edit_grid.addWidget(self.path_label, 1, 0, 1, 1)

        self.name_line_edit = ConanRefLineEdit(Dialog)
        self.name_line_edit.setObjectName(u"name_line_edit")

        self.edit_grid.addWidget(self.name_line_edit, 0, 1, 1, 1)

        self.output_folder_line_edit = QLineEdit(Dialog)
        self.output_folder_line_edit.setObjectName(u"output_folder_line_edit")

        self.edit_grid.addWidget(self.output_folder_line_edit, 2, 1, 1, 1)

        self.output_folder_browse_button = QPushButton(Dialog)
        self.output_folder_browse_button.setObjectName(u"output_folder_browse_button")
        self.output_folder_browse_button.setMaximumSize(QSize(50, 16777215))

        self.edit_grid.addWidget(self.output_folder_browse_button, 2, 2, 1, 1)

        self.edit_grid.setRowStretch(0, 1)

        self.verticalLayout.addLayout(self.edit_grid)

        self.button_box = QDialogButtonBox(Dialog)
        self.button_box.setObjectName(u"button_box")
        sizePolicy3 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        sizePolicy3.setHorizontalStretch(1)
        sizePolicy3.setVerticalStretch(1)
        sizePolicy3.setHeightForWidth(self.button_box.sizePolicy().hasHeightForWidth())
        self.button_box.setSizePolicy(sizePolicy3)
        self.button_box.setOrientation(Qt.Horizontal)
        self.button_box.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.button_box)


        self.retranslateUi(Dialog)
        self.button_box.rejected.connect(Dialog.reject)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Edit Remote", None))
        self.output_folder_label.setText(QCoreApplication.translate("Dialog", u"Output Folder", None))
        self.path_browse_button.setText(QCoreApplication.translate("Dialog", u"...", None))
        self.name_label.setText(QCoreApplication.translate("Dialog", u"Reference", None))
        self.path_label.setText(QCoreApplication.translate("Dialog", u"Path", None))
        self.output_folder_browse_button.setText(QCoreApplication.translate("Dialog", u"...", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'conan_cache_cleanup.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QDialog, QDialogButtonBox,
    QFrame, QHBoxLayout, QHeaderView, QLabel,
    QSizePolicy, QTreeWidget, QTreeWidgetItem, QVBoxLayout,
    QWidget)

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(524, 307)
        Dialog.setSizeGripEnabled(True)
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.frame = QFrame(Dialog)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.Shape.StyledPanel)
        self.frame.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout = QHBoxLayout(self.frame)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.question_label = QLabel(self.frame)
        self.question_label.setObjectName(u"question_label")
        self.question_label.setLineWidth(0)

        self.horizontalLayout.addWidget(self.question_label)

        self.icon = QLabel(self.frame)
        self.icon.setObjectName(u"icon")

        self.horizontalLayout.addWidget(self.icon)

        self.horizontalLayout.setStretch(0, 1)

        self.verticalLayout.addWidget(self.frame)

        self.cleanup_tree_widget = QTreeWidget(Dialog)
        self.cleanup_tree_widget.setObjectName(u"cleanup_tree_widget")

        self.verticalLayout.addWidget(self.cleanup_tree_widget)

        self.button_box = QDialogButtonBox(Dialog)
        self.button_box.setObjectName(u"button_box")
        self.button_box.setOrientation(Qt.Orientation.Horizontal)
        self.button_box.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)

        self.verticalLayout.addWidget(self.button_box)


        self.retranslateUi(Dialog)
        self.button_box.accepted.connect(Dialog.accept)
        self.button_box.rejected.connect(Dialog.reject)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Delete folders", None))
        self.question_label.setText(QCoreApplication.translate("Dialog", u"TextLabel", None))
        self.icon.setText("")
        ___qtreewidgetitem = self.cleanup_tree_widget.headerItem()
        ___qtreewidgetitem.setText(1, QCoreApplication.translate("Dialog", u"Size", None));
        ___qtreewidgetitem.setText(0, QCoreApplication.translate("Dialog", u"Conan Reference", None));
    # retranslateUi




""" Common ui classes, and functions """

from pathlib import Path

from PySide6.QtGui import QFont, QFontMetrics

import conan_explorer.app as app

# using global module pattern
from conan_explorer.app.system import execute_cmd, open_file
from conan_explorer.conan_wrapper.types import ConanRef
from conan_explorer.settings import FILE_EDITOR_EXECUTABLE, FONT_SIZE

from .icon import (
    draw_svg_with_color,
    extract_icon,
    get_icon_from_image_file,
    get_inverted_asset_image,
    get_platform_icon,
)
from .logger import init_qt_logger, remove_qt_logger
from .model import FileSystemModel, TreeModel, TreeModelItem, re_register_signal
from .syntax_highlighting import ConfigHighlighter
from .theming import (
    CanSetIconWidgetProtocol,
    CanSetPixmapWidgetProtocol,
    ThemedWidget,
    get_asset_image_path,
    get_gui_dark_mode,
    get_gui_style,
    get_themed_asset_icon,
)


def measure_font_width(text: str) -> int:
    """ Return the width of a text in pixels woth the current font and default fontsize """
    fs = app.active_settings.get_int(FONT_SIZE)
    font = QFont()
    font.setPointSize(fs)
    fm = QFontMetrics(font)
    return fm.horizontalAdvance(text)


def show_conanfile(conan_ref: str):
    conanfile = app.conan_api.get_conanfile_path(ConanRef.loads(conan_ref))
    editor = app.active_settings.get_string(FILE_EDITOR_EXECUTABLE)
    if not editor or not Path(editor).exists():
        open_file(conanfile)
        return
    execute_cmd([editor, str(conanfile)], False)



from ast import Str
from typing import List, Optional

from PySide6.QtCore import QModelIndex, QPersistentModelIndex, Qt
from PySide6.QtGui import QFont
from typing_extensions import override

import conan_explorer.app as app  # using global module pattern
from conan_explorer.app.logger import Logger
from conan_explorer.conan_wrapper.types import ConanException, Remote
from conan_explorer.ui.common import TreeModel, TreeModelItem


class RemotesModelItem(TreeModelItem, Remote):
    def __init__(self, remote: Remote, user: str, auth: bool, parent=None, lazy_loading=False):
        TreeModelItem.__init__(self, [user, str(auth)], parent, lazy_loading=lazy_loading)
        Remote.__init__(self, remote.name, remote.url, remote.verify_ssl, remote.disabled)

    @override
    def data(self, column):
        if column == 0:
            return self.name
        elif column == 1:
            return self.url
        elif column == 2:
            return str(self.verify_ssl)
        elif column == 3:
            return self.item_data[0]
        elif column == 4:
            return self.item_data[1]
        elif column == 5:
            return str(self.disabled)

    @property
    def user(self) -> str:
        return self.item_data[0]

    @user.setter
    def user(self, value: str):
        self.item_data[0] = value

    @property
    def auth(self) -> bool:
        return bool(self.item_data[1])

    @auth.setter
    def auth(self, value: bool):
        self.item_data[1] = str(value)


class RemotesTableModel(TreeModel):
    """
    Remotes displayed as a table with detail info - implemented as a tree with one level:
    this is necessary, because list model cannot have a header and multiple columns and a
    table looks like an ugly Excel clone
    """

    def __init__(self, *args, **kwargs):
        super(RemotesTableModel, self).__init__(checkable=True, *args, **kwargs)
        self.root_item = TreeModelItem(["Name", "URL", "SSL", "User", "Authenticated"])
        self.setup_model_data()

    def setup_model_data(self):
        self.clear_items()
        self.beginResetModel()
        for remote in app.conan_api.get_remotes(include_disabled=True):
            user_name = ""
            auth = False
            try:
                user_name, auth = app.conan_api.get_remote_user_info(remote.name)
            # Throws an error on older conan version, if a remote is disabled
            except ConanException:
                Logger().debug("", exc_info=True)
            remote_item = RemotesModelItem(remote, user_name, auth, self.root_item)
            self.root_item.append_child(remote_item)
        self.endResetModel()

    @override
    def data(self, index, role):
        if not index.isValid():
            return None
        item: RemotesModelItem = index.internalPointer()  # type: ignore
        if role == Qt.ItemDataRole.DisplayRole:
            try:
                return item.data(index.column())
            except Exception:
                return ""

        if isinstance(item, RemotesModelItem):
            if role == Qt.ItemDataRole.FontRole and item.disabled:
                font = QFont()
                font.setItalic(True)
                return font

        return None

    @override
    def rowCount(self, parent=None):  # TODO really?
        return self.root_item.child_count()

    def items(self) -> List[RemotesModelItem]:
        return self.root_item.child_items  # type: ignore

    def add(self, remote: Remote):
        app.conan_api.add_remote(remote.name, remote.url, remote.verify_ssl)
        super().add_item(RemotesModelItem(remote, "", False))

    def remove(self, remote_name: str):
        app.conan_api.remove_remote(remote_name)
        index = self.get_index_from_ref(remote_name)
        item: RemotesModelItem = index.internalPointer()  # type: ignore
        super().remove_item(item)
        return True

    def rename(self, remote: Remote, new_name):
        app.conan_api.rename_remote(remote.name, new_name)
        index = self.get_index_from_ref(remote.name)
        item: RemotesModelItem = index.internalPointer()  # type: ignore
        item.name = new_name

    def update(self, remote: Remote):
        app.conan_api.update_remote(
            remote.name, remote.url, remote.verify_ssl, remote.disabled, None
        )
        index = self.get_index_from_ref(remote.name)
        item: RemotesModelItem = index.internalPointer()  # type: ignore
        # copy all possible props
        item.url = remote.url
        item.verify_ssl = remote.verify_ssl
        item.disabled = remote.disabled

    def update_login_info(self, remote_name: str, user: str, pwd: str):
        app.conan_api.login_remote(remote_name, user, pwd)
        index = self.get_index_from_ref(remote_name)
        item: RemotesModelItem = index.internalPointer()  # type: ignore
        item.auth = True
        item.user = user

    def save(self):
        """Update every remote with new index and then save to conan remotes file"""
        i = 0
        for remote in self.items():
            app.conan_api.update_remote(
                remote.name, remote.url, remote.verify_ssl, remote.disabled, i
            )
            i += 1

    @override
    def moveRow(
        self,
        source_parent: "QModelIndex | QPersistentModelIndex",
        source_row: int,
        destination_parent: QModelIndex,
        destination_child: int,
    ) -> bool:
        item_to_move = self.items()[source_row]
        self.items().insert(destination_child, item_to_move)
        if source_row < destination_child:
            self.items().pop(source_row)
        else:
            self.items().pop(source_row + 1)
        self.save()  # autosave
        return super().moveRow(source_parent, source_row, destination_parent, destination_child)

    def get_index_from_ref(self, conan_ref: str) -> Optional[QModelIndex]:
        for ref in self.items():
            if ref.name == conan_ref:
                return self.get_index_from_item(ref)
        return None



from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Tuple

from PySide6.QtCore import (
    QItemSelectionModel,
    QMimeData,
    QModelIndex,
    QObject,
    Qt,
    QUrl,
    SignalInstance,
)
from PySide6.QtWidgets import (
    QApplication,
    QDialog,
    QListWidgetItem,
    QMessageBox,
    QStyle,
    QTextBrowser,
    QTreeView,
)

import conan_explorer.app as app
from conan_explorer.app.base_ui.loading import LoaderGui
from conan_explorer.app.logger import Logger
from conan_explorer.app.system import (
    calc_paste_same_dir_name,
    copy_path_with_overwrite,
    delete_path,
    execute_cmd,
    open_cmd_in_path,
    open_in_file_manager,
    run_file,
)
from conan_explorer.conan_wrapper.types import ConanPkg, ConanRef
from conan_explorer.settings import FILE_EDITOR_EXECUTABLE
from conan_explorer.ui.common import re_register_signal
from conan_explorer.ui.dialogs import QuestionWithItemListDialog
from conan_explorer.ui.views import AppGridView
from conan_explorer.ui.views.app_grid import UiAppLinkConfig

from .file_model import CalFileSystemModel
from .sel_model import PkgSelectionType

if TYPE_CHECKING:
    from conan_explorer.ui.fluent_window import FluentWindow
    from conan_explorer.ui.main_window import BaseSignals

    from .package_explorer import LocalConanPackageExplorer

NEW_FOLDER_NAME = "New folder"


class PackageFileExplorerController(QObject):
    def __init__(
        self,
        parent: "LocalConanPackageExplorer",
        view: QTreeView,
        pkg_path_label: QTextBrowser,
        cut_files_reset: SignalInstance,
        base_signals: "BaseSignals",
        page_widgets: "FluentWindow.PageStore",
    ):
        super().__init__(parent)
        self._model: Optional[CalFileSystemModel] = None
        self._page_widgets = page_widgets
        self._view = view
        self._pkg_path_label = pkg_path_label
        self.cut_files_reset = cut_files_reset
        self._type = PkgSelectionType.ref
        base_signals.conan_pkg_removed.connect(self.on_conan_pkg_removed)
        self.cut_files_reset.connect(self._disable_cut_selection)

        self._current_ref: Optional[str] = None  # loaded conan ref
        self._current_pkg: Optional[ConanPkg] = None  # loaded conan pkg info

    def get_conan_pkg_info(self):
        return self._current_pkg

    def get_conan_pkg_type(self):
        return self._type

    def on_pkg_selection_change(
        self, conan_ref: str, pkg_info: ConanPkg, type: PkgSelectionType
    ):
        """Change folder in file view"""
        # add try-except
        self._current_ref = conan_ref
        self._current_pkg = pkg_info
        try:
            if type == PkgSelectionType.editable:
                pkg_path = app.conan_api.get_editables_package_path(ConanRef.loads(conan_ref))
                if pkg_path.is_file():
                    pkg_path = pkg_path.parent
            elif type == PkgSelectionType.export:
                pkg_path = app.conan_api.get_export_folder(ConanRef.loads(conan_ref))
            else:
                pkg_path = app.conan_api.get_package_folder(
                    ConanRef.loads(conan_ref), pkg_info.get("id", "")
                )
        except Exception as e:
            Logger().error("Cannot change to package: %s", str(e))
            return

        if not pkg_path.exists():
            Logger().warning(
                f"Can't find package path for {conan_ref} and {str(pkg_info)} for File View"
            )
            return

        if self._model:
            if pkg_path == Path(self._model.rootPath()):
                return
        else:  # initialize once - otherwise this causes performance issues
            self._model = CalFileSystemModel()
            self._view.setModel(self._model)

        self._model.setRootPath(str(pkg_path))
        self._model.sort(0, Qt.SortOrder.AscendingOrder)
        self._model.setReadOnly(False)
        self._view.setRootIndex(self._model.index(str(pkg_path)))
        self._view.setColumnHidden(2, True)  # file type
        self._model.layoutChanged.connect(self.resize_file_columns)
        self._view.header().setSortIndicator(0, Qt.SortOrder.AscendingOrder)
        # disable edit on double click, since we want to open
        re_register_signal(self._view.doubleClicked, self.on_file_double_click)  # type: ignore

        self._view.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.resize_file_columns()

    def on_conan_pkg_removed(self, conan_ref: str, pkg_id: str):
        """Slot for on_conan_pkg_removed signal."""
        # clear file view if this pkg is selected
        own_id = pkg_id
        if self._current_pkg:
            own_id = self._current_pkg.get("id", "")
        if self._current_ref == conan_ref:
            delete = False
            if not pkg_id:
                delete = True
            elif pkg_id and pkg_id == own_id:
                delete = True
            if not delete:
                return
            self.close_files_view()
            self.parent().tab_close_requested(self)  # type: ignore

    def close_files_view(self):
        """Reset and clear up file view"""
        if self._model:
            self._model.deleteLater()
        self._model = None  # type: ignore
        self._current_ref = ""
        self._current_pkg = None
        self._pkg_path_label.setText("")
        self._view.setModel(None)  # type: ignore
        self._pkg_path_label.setText("")

    def on_file_double_click(self, model_index):
        file_path = Path(model_index.model().fileInfo(model_index).absoluteFilePath())
        run_file(file_path, True, args="")

    def on_copy_file_as_path(self) -> str:
        files = self.get_selected_pkg_paths()
        file_paths = "\n".join(files)
        QApplication.clipboard().setText(file_paths)
        return file_paths

    def on_open_terminal_in_dir(self) -> int:
        selected_path = Path(self.get_selected_pkg_paths()[0])  # TODO!
        if selected_path.is_file():
            selected_path = selected_path.parent
        return open_cmd_in_path(selected_path)

    def on_edit_file(self):
        selected_paths = self.get_selected_pkg_paths()
        if len(selected_paths) != 1:
            return
        for selected_path in selected_paths:
            if not Path(selected_path).is_file():
                continue
            editor = app.active_settings.get_string(FILE_EDITOR_EXECUTABLE)
            if not editor:
                Logger().info("No editor set up! Please configure it in the settings menu.")
                return
            execute_cmd([editor, selected_path], False)

    def on_file_delete(self):
        selected_paths = self.get_selected_pkg_paths()
        msg = QMessageBox(parent=self._view)
        msg.setWindowTitle("Delete file")
        msg.setText("Are you sure, you want to permanently delete this file(s)/folder(s)?\t")
        msg.setStandardButtons(
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.Cancel
        )
        msg.setIcon(QMessageBox.Icon.Warning)
        reply = msg.exec()
        if reply != QMessageBox.StandardButton.Yes:
            return
        for path in selected_paths:
            # TODO loading gui
            delete_path(Path(path))

    def on_files_copy(self) -> List[QUrl]:
        if not self._model:
            return []
        files = self.get_selected_pkg_paths()
        if not files:
            return []
        data = QMimeData()
        urls = []
        for file in files:
            urls.append(QUrl.fromLocalFile(file))
        self.cut_files_reset.emit()
        self._view.repaint()
        data.setUrls(urls)

        QApplication.clipboard().setMimeData(data)
        return urls

    def on_files_cut(self) -> List[QUrl]:
        if not self._model:
            return []
        files = self.get_selected_pkg_paths()
        if not files:
            return []
        urls = []
        for file in files:
            urls.append(QUrl.fromLocalFile(file))
        data = QMimeData()
        data.setUrls(urls)
        data.setProperty("action", "cut")
        self._model.clear_all_disabled_items()  # type: ignore
        self.cut_files_reset.emit()

        self._model.add_disabled_items(files)  # type: ignore
        self._view.repaint()

        QApplication.clipboard().setMimeData(data)
        return urls

    def on_item_rename(self):
        file_view_indexes = self._get_pkg_file_source_items()
        if not file_view_indexes:
            return None
        # for multiselect: edit last item
        self._view.edit(file_view_indexes[-1])
        return file_view_indexes

    def on_files_paste(self):
        data = QApplication.clipboard().mimeData()
        if not data or not data.hasUrls():
            return
        urls = data.urls()
        # determine destination path
        sel_item_paths = self.get_selected_pkg_paths()
        if not sel_item_paths:
            return
        # determine destination path
        dst_dir_path = Path(sel_item_paths[0])

        # on multiselect it will paste once in the parent dir of the selected items
        if len(sel_item_paths) > 1:
            dst_dir_path = dst_dir_path.parent

        elif not self._is_selected_file_item_expanded():
            dst_dir_path = dst_dir_path.parent if dst_dir_path.is_file() else dst_dir_path

        # all of them should be true
        cut = True if data.property("action") == "cut" else False
        files_paths: List[Tuple[Path, Path]] = []
        overwrites: List[Path] = []
        for url in urls:
            # try to copy
            if not url.isLocalFile():
                continue
            source_path = Path(url.toLocalFile())
            target_path = Path(dst_dir_path, url.fileName())
            files_paths.append((source_path, target_path))
            if target_path.exists():
                overwrites.append(target_path)

        # Show dialog which files should be overwritten
        if overwrites:
            non_overwrites_paths = self.show_overwrite_dialog(overwrites)
            # remove selected_overwrites from files_paths
            files_paths = list(
                filter(
                    lambda files_path: files_path[1] not in non_overwrites_paths, files_paths
                )
            )
        if not files_paths:
            return
        loader = LoaderGui(None)

        def paste_files(self):
            for source_path, target_path in files_paths:
                self.paste_path(source_path, target_path, cut)

        loader.load(self._view, paste_files, (self,))
        self.cut_files_reset.emit()
        self._disable_cut_selection()

    def _disable_cut_selection(self):
        if self.sender() == self:
            return
        self._model.clear_all_disabled_items()  # type: ignore
        self._view.repaint()

    def paste_path(self, src: Path, dst: Path, cut=False):
        if src == dst and cut:
            return  # do nothing
        if src == dst:  # same target -> create numbered copies
            new_dst = calc_paste_same_dir_name(dst)
            copy_path_with_overwrite(src, new_dst)
        else:
            copy_path_with_overwrite(src, dst)
            if cut:
                delete_path(src)

    def show_overwrite_dialog(self, overwrites: List[Path]):
        non_overwrites = []
        dialog = QuestionWithItemListDialog(
            self._view, QStyle.StandardPixmap.SP_MessageBoxWarning
        )
        dialog.setWindowTitle("Overwrite file/folder")
        dialog.set_question_text("Are you sure, you want to overwrite this file/folder?\t")
        for path in overwrites:
            list_item = QListWidgetItem(str(path))
            list_item.setCheckState(Qt.CheckState.Checked)
            dialog.item_list_widget.addItem(list_item)

        reply = dialog.exec()
        if reply == QDialog.DialogCode.Accepted:
            # get user DEselection
            for list_row in range(dialog.item_list_widget.count()):
                list_item = dialog.item_list_widget.item(list_row)
                if list_item.checkState() != Qt.CheckState.Unchecked:
                    continue
                non_overwrites.append(Path(list_item.text()))
            return non_overwrites
        else:
            return overwrites  # early return to not delete path on cut and cancel

    def on_add_app_link_from_file(self):
        selected_paths = self.get_selected_pkg_paths()
        if len(selected_paths) != 1:
            return
        file_path = Path(selected_paths[0])
        if not file_path.is_file():
            Logger().error("Please select a file, not a directory!")
            return
        if not self._current_ref:
            return
        conan_ref = ConanRef.loads(self._current_ref)
        # determine relpath from package
        pkg_info = self._current_pkg
        if not pkg_info:
            Logger().error("Error on trying to adding Applink: No package info found.")
            return
        pkg_path = app.conan_api.get_package_folder(conan_ref, pkg_info.get("id", ""))
        rel_path = file_path.relative_to(pkg_path).as_posix()  # no \ for UI config

        app_config = UiAppLinkConfig(
            name="NewLink",
            conan_ref=str(self._current_ref),
            executable=str(rel_path),
            conan_options=pkg_info.get("options", {}),
        )
        self._page_widgets.get_page_by_type(AppGridView).open_new_app_dialog_from_extern(
            app_config
        )

    def on_new_folder(self, model_index):
        request_file_path = Path(self.get_selected_pkg_paths()[-1])
        target_file_path = request_file_path
        if request_file_path.is_file():
            target_file_path = request_file_path.parent

        target_folder = target_file_path / NEW_FOLDER_NAME
        target_folder.mkdir(exist_ok=True)
        # it does not seem to be necessary to wait for creation
        if self.select_file_item(str(target_folder)).isValid():
            self.on_item_rename()

    def on_open_file_in_file_manager(self, model_index):
        file_path = Path(self.get_selected_pkg_paths()[-1])
        open_in_file_manager(file_path)

    def _get_pkg_file_source_items(self) -> List[QModelIndex]:
        indexes = []
        for index in self._view.selectedIndexes():
            if index.column() == 0:  # we only need a row once
                indexes.append(index)
        return indexes

    def select_file_item(self, file_to_select: str) -> QModelIndex:
        if not self._model:
            return QModelIndex()
        try:
            sel_idx = self._model.index(file_to_select, 0)
            self._view.selectionModel().select(
                sel_idx, QItemSelectionModel.SelectionFlag.ClearAndSelect
            )
        except Exception:
            Logger().debug("Can't select file %s", file_to_select, exc_info=True)
            return QModelIndex()
        return sel_idx

    def get_selected_pkg_paths(self) -> List[str]:
        file_paths: List[str] = []
        if not self._model:
            return file_paths
        file_view_indexes = self._get_pkg_file_source_items()
        # if nothing selected return root
        if not file_view_indexes:
            return [self._model.rootPath()]
        for file_view_index in file_view_indexes:
            file_paths.append(self._model.fileInfo(file_view_index).absoluteFilePath())
        return file_paths

    def _is_selected_file_item_expanded(self):
        file_view_indexes = self._get_pkg_file_source_items()
        # if nothing selected return root
        if len(file_view_indexes) != 1:
            return False
        return self._view.isExpanded(file_view_indexes[0])

    def resize_file_columns(self):
        if not self._view:
            return
        for i in reversed(range(3 - 1)):
            self._view.resizeColumnToContents(i)



from PySide6.QtWidgets import QLineEdit

from conan_explorer.ui.common import get_themed_asset_icon


class PasswordLineEdit(QLineEdit):
    def __init__(self, parent):
        super().__init__(parent)
        # hide per default
        self.setEchoMode(self.EchoMode.Password)

        self.show_icon = get_themed_asset_icon("icons/show.svg")
        self.hide_icon = get_themed_asset_icon("icons/hide.svg")

        self.toggle_show_pw_action = self.addAction(
            self.show_icon, self.ActionPosition.TrailingPosition
        )
        self.toggle_show_pw_action.triggered.connect(self.on_toggle_show_password)
        self._pw_shown = False

    def on_toggle_show_password(self):
        if self._pw_shown:
            self.setEchoMode(self.EchoMode.Password)
            self._pw_shown = False
            self.toggle_show_pw_action.setIcon(self.show_icon)
        else:
            self.setEchoMode(self.EchoMode.Normal)
            self._pw_shown = True
            self.toggle_show_pw_action.setIcon(self.hide_icon)



import os
from pathlib import Path

from conan_explorer import DEBUG_LEVEL, INVALID_PATH

from .fluent_window import FluentWindow
from .main_window import BaseSignals
from .plugin import PluginDescription, PluginFile, PluginInterfaceV1

__all__ = [
    "BaseSignals",
    "FluentWindow",
    "PluginInterfaceV1",
    "PluginFile",
    "PluginDescription",
]


def compile_ui_file_if_newer(ui_file: Path):
    from conan_explorer.app.logger import Logger

    py_ui_file = ui_file.parent / (ui_file.stem + "_ui.py")
    if py_ui_file.exists() and py_ui_file.stat().st_mtime > ui_file.stat().st_mtime:
        return
    Logger().debug("Converting " + str(py_ui_file))
    os.system(f"pyside6-uic -o {str(py_ui_file)} {str(ui_file)}")


# compile uic files, if needed
if DEBUG_LEVEL > 0:
    current_dir = Path(__file__).parent
    for ui_file in current_dir.glob("**/*.ui"):
        py_ui_file = Path(INVALID_PATH)
        try:
            compile_ui_file_if_newer(ui_file)
        except Exception as e:
            from conan_explorer.app.logger import Logger

            Logger().warning(f"Can't convert {str(py_ui_file)}: {str(e)}")



import ctypes
import platform


def gen_obj_name(name: str) -> str:
    """Generates an object name from a menu title or name
    (spaces to underscores and lowercase)"""
    return name.replace(" ", "_").lower()


def get_display_scaling():
    if platform.system() == "Windows":
        return ctypes.windll.shcore.GetScaleFactorForDevice(0) / 100
    else:  # TODO not yet implemented for Linux
        return 2.2


LEFT_MENU_MIN_WIDTH = 80
LEFT_MENU_MAX_WIDTH = int(310 + 20 * (2 / get_display_scaling()))
RIGHT_MENU_MIN_WIDTH = 0
RIGHT_MENU_MAX_WIDTH = int(160 + 200 * (2 / get_display_scaling()))


from .fluent_window import FluentWindow  # noqa
from .side_menu import SideSubMenu  # noqa

__all__ = ["FluentWindow", "SideSubMenu", "get_display_scaling", "gen_obj_name"]



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'fluent_window.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QAbstractScrollArea, QApplication, QFrame,
    QGridLayout, QHBoxLayout, QLabel, QLayout,
    QLineEdit, QListView, QMainWindow, QPushButton,
    QScrollArea, QSizePolicy, QSpacerItem, QSplitter,
    QStackedWidget, QTextBrowser, QTextEdit, QVBoxLayout,
    QWidget)

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(MainWindow.sizePolicy().hasHeightForWidth())
        MainWindow.setSizePolicy(sizePolicy)
        MainWindow.setMinimumSize(QSize(400, 400))
        palette = QPalette()
        brush = QBrush(QColor(160, 160, 160, 255))
        brush.setStyle(Qt.SolidPattern)
        palette.setBrush(QPalette.Active, QPalette.Highlight, brush)
        palette.setBrush(QPalette.Inactive, QPalette.Highlight, brush)
        brush1 = QBrush(QColor(0, 120, 215, 255))
        brush1.setStyle(Qt.SolidPattern)
        palette.setBrush(QPalette.Disabled, QPalette.Highlight, brush1)
        MainWindow.setPalette(palette)
        font = QFont()
        font.setStyleStrategy(QFont.PreferAntialias)
        MainWindow.setFont(font)
        MainWindow.setAcceptDrops(True)
        icon = QIcon()
        icon.addFile(u":/icons/icon.ico", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        MainWindow.setWindowIcon(icon)
        self.central_widget = QWidget(MainWindow)
        self.central_widget.setObjectName(u"central_widget")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy1.setHorizontalStretch(1)
        sizePolicy1.setVerticalStretch(1)
        sizePolicy1.setHeightForWidth(self.central_widget.sizePolicy().hasHeightForWidth())
        self.central_widget.setSizePolicy(sizePolicy1)
        self.central_widget_layout = QVBoxLayout(self.central_widget)
        self.central_widget_layout.setSpacing(0)
        self.central_widget_layout.setObjectName(u"central_widget_layout")
        self.central_widget_layout.setContentsMargins(2, 0, 1, 1)
        self.main_frame = QFrame(self.central_widget)
        self.main_frame.setObjectName(u"main_frame")
        self.main_frame.setFrameShape(QFrame.Shape.NoFrame)
        self.main_frame.setFrameShadow(QFrame.Shadow.Raised)
        self.main_frame.setLineWidth(0)
        self.main_frame_layout = QGridLayout(self.main_frame)
        self.main_frame_layout.setSpacing(0)
        self.main_frame_layout.setObjectName(u"main_frame_layout")
        self.main_frame_layout.setContentsMargins(0, 0, 3, 1)
        self.top_frame = QFrame(self.main_frame)
        self.top_frame.setObjectName(u"top_frame")
        self.top_frame.setMinimumSize(QSize(0, 55))
        self.top_frame.setMaximumSize(QSize(16777215, 70))
        self.top_frame.setLineWidth(0)
        self.top_frame_layout = QHBoxLayout(self.top_frame)
        self.top_frame_layout.setSpacing(0)
        self.top_frame_layout.setObjectName(u"top_frame_layout")
        self.top_frame_layout.setContentsMargins(1, 0, 0, 0)
        self.top_center_frame = QFrame(self.top_frame)
        self.top_center_frame.setObjectName(u"top_center_frame")
        self.top_center_frame.setMinimumSize(QSize(55, 0))
        self.top_center_frame.setMaximumSize(QSize(16777214, 100))
        self.top_center_frame_layout = QHBoxLayout(self.top_center_frame)
        self.top_center_frame_layout.setObjectName(u"top_center_frame_layout")
        self.top_center_frame_layout.setContentsMargins(5, 1, 0, 2)
        self.page_title = QLabel(self.top_center_frame)
        self.page_title.setObjectName(u"page_title")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.MinimumExpanding)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.page_title.sizePolicy().hasHeightForWidth())
        self.page_title.setSizePolicy(sizePolicy2)
        font1 = QFont()
        font1.setPointSize(24)
        font1.setStyleStrategy(QFont.PreferAntialias)
        self.page_title.setFont(font1)
        self.page_title.setScaledContents(True)
        self.page_title.setAlignment(Qt.AlignmentFlag.AlignBottom|Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft)
        self.page_title.setMargin(0)

        self.top_center_frame_layout.addWidget(self.page_title)

        self.top_center_right_frame = QFrame(self.top_center_frame)
        self.top_center_right_frame.setObjectName(u"top_center_right_frame")
        sizePolicy3 = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Preferred)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.top_center_right_frame.sizePolicy().hasHeightForWidth())
        self.top_center_right_frame.setSizePolicy(sizePolicy3)
        self.top_center_right_frame.setMinimumSize(QSize(0, 0))
        self.gridLayout = QGridLayout(self.top_center_right_frame)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(9, 0, 3, 0)
        self.window_btns_frame = QFrame(self.top_center_right_frame)
        self.window_btns_frame.setObjectName(u"window_btns_frame")
        self.window_btns_frame.setMinimumSize(QSize(0, 45))
        self.window_btns_frame.setMaximumSize(QSize(165, 60))
        self.window_btns_frame.setLayoutDirection(Qt.LayoutDirection.RightToLeft)
        self.window_btns_frame.setFrameShape(QFrame.Shape.NoFrame)
        self.horizontalLayout_4 = QHBoxLayout(self.window_btns_frame)
        self.horizontalLayout_4.setSpacing(0)
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.horizontalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.close_button = QPushButton(self.window_btns_frame)
        self.close_button.setObjectName(u"close_button")
        self.close_button.setMinimumSize(QSize(55, 35))
        self.close_button.setFlat(True)

        self.horizontalLayout_4.addWidget(self.close_button)

        self.restore_max_button = QPushButton(self.window_btns_frame)
        self.restore_max_button.setObjectName(u"restore_max_button")
        self.restore_max_button.setMinimumSize(QSize(55, 35))
        self.restore_max_button.setFlat(True)

        self.horizontalLayout_4.addWidget(self.restore_max_button)

        self.minimize_button = QPushButton(self.window_btns_frame)
        self.minimize_button.setObjectName(u"minimize_button")
        self.minimize_button.setMinimumSize(QSize(55, 35))
        self.minimize_button.setFlat(True)

        self.horizontalLayout_4.addWidget(self.minimize_button)


        self.gridLayout.addWidget(self.window_btns_frame, 0, 1, 1, 1)

        self.top_center_vspacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.top_center_vspacer, 1, 1, 1, 1)

        self.top_center_search_frame = QFrame(self.top_center_right_frame)
        self.top_center_search_frame.setObjectName(u"top_center_search_frame")
        self.top_center_search_frame.setLayoutDirection(Qt.LayoutDirection.RightToLeft)
        self.verticalLayout = QVBoxLayout(self.top_center_search_frame)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, -1, 0, -1)
        self.search_bar_line_edit = QLineEdit(self.top_center_search_frame)
        self.search_bar_line_edit.setObjectName(u"search_bar_line_edit")
        self.search_bar_line_edit.setMaximumSize(QSize(400, 16777215))
        self.search_bar_line_edit.setLayoutDirection(Qt.LayoutDirection.RightToLeft)
        self.search_bar_line_edit.setClearButtonEnabled(True)

        self.verticalLayout.addWidget(self.search_bar_line_edit)


        self.gridLayout.addWidget(self.top_center_search_frame, 0, 0, 1, 1)


        self.top_center_frame_layout.addWidget(self.top_center_right_frame)


        self.top_frame_layout.addWidget(self.top_center_frame)


        self.main_frame_layout.addWidget(self.top_frame, 0, 1, 1, 1)

        self.center_frame = QFrame(self.main_frame)
        self.center_frame.setObjectName(u"center_frame")
        self.center_frame.setFrameShape(QFrame.Shape.NoFrame)
        self.center_frame.setFrameShadow(QFrame.Shadow.Raised)
        self.center_frame_layout = QHBoxLayout(self.center_frame)
        self.center_frame_layout.setSpacing(0)
        self.center_frame_layout.setObjectName(u"center_frame_layout")
        self.center_frame_layout.setSizeConstraint(QLayout.SizeConstraint.SetDefaultConstraint)
        self.center_frame_layout.setContentsMargins(0, 0, 2, 0)
        self.page_frame = QFrame(self.center_frame)
        self.page_frame.setObjectName(u"page_frame")
        self.page_frame.setFrameShape(QFrame.Shape.NoFrame)
        self.page_frame.setFrameShadow(QFrame.Shadow.Raised)
        self.page_frame_layout = QVBoxLayout(self.page_frame)
        self.page_frame_layout.setSpacing(0)
        self.page_frame_layout.setObjectName(u"page_frame_layout")
        self.page_frame_layout.setSizeConstraint(QLayout.SizeConstraint.SetMinAndMaxSize)
        self.page_frame_layout.setContentsMargins(0, 0, 0, 0)
        self.content_footer_splitter = QSplitter(self.page_frame)
        self.content_footer_splitter.setObjectName(u"content_footer_splitter")
        self.content_footer_splitter.setOrientation(Qt.Orientation.Vertical)
        self.content_footer_splitter.setChildrenCollapsible(False)
        self.content_frame = QFrame(self.content_footer_splitter)
        self.content_frame.setObjectName(u"content_frame")
        sizePolicy4 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy4.setHorizontalStretch(0)
        sizePolicy4.setVerticalStretch(0)
        sizePolicy4.setHeightForWidth(self.content_frame.sizePolicy().hasHeightForWidth())
        self.content_frame.setSizePolicy(sizePolicy4)
        self.content_frame.setMinimumSize(QSize(0, 300))
        self.content_frame.setFrameShape(QFrame.Shape.NoFrame)
        self.content_frame.setFrameShadow(QFrame.Shadow.Raised)
        self.content_frame_layout = QVBoxLayout(self.content_frame)
        self.content_frame_layout.setSpacing(0)
        self.content_frame_layout.setObjectName(u"content_frame_layout")
        self.content_frame_layout.setSizeConstraint(QLayout.SizeConstraint.SetMinAndMaxSize)
        self.content_frame_layout.setContentsMargins(2, 0, 2, 1)
        self.page_stacked_widget = QStackedWidget(self.content_frame)
        self.page_stacked_widget.setObjectName(u"page_stacked_widget")
        sizePolicy4.setHeightForWidth(self.page_stacked_widget.sizePolicy().hasHeightForWidth())
        self.page_stacked_widget.setSizePolicy(sizePolicy4)
        self.page = QWidget()
        self.page.setObjectName(u"page")
        self.page_layout = QVBoxLayout(self.page)
        self.page_layout.setObjectName(u"page_layout")
        self.page_layout.setSizeConstraint(QLayout.SizeConstraint.SetMaximumSize)
        self.page_layout.setContentsMargins(2, 0, 0, 0)
        self.listView_2 = QListView(self.page)
        self.listView_2.setObjectName(u"listView_2")
        sizePolicy4.setHeightForWidth(self.listView_2.sizePolicy().hasHeightForWidth())
        self.listView_2.setSizePolicy(sizePolicy4)
        self.listView_2.setMinimumSize(QSize(0, 400))
        self.listView_2.setFrameShape(QFrame.Shape.NoFrame)
        self.listView_2.setFrameShadow(QFrame.Shadow.Plain)
        self.listView_2.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.listView_2.setHorizontalScrollMode(QAbstractItemView.ScrollMode.ScrollPerPixel)

        self.page_layout.addWidget(self.listView_2)

        self.page_stacked_widget.addWidget(self.page)

        self.content_frame_layout.addWidget(self.page_stacked_widget)

        self.content_footer_splitter.addWidget(self.content_frame)
        self.page_footer = QFrame(self.content_footer_splitter)
        self.page_footer.setObjectName(u"page_footer")
        sizePolicy5 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Ignored)
        sizePolicy5.setHorizontalStretch(0)
        sizePolicy5.setVerticalStretch(0)
        sizePolicy5.setHeightForWidth(self.page_footer.sizePolicy().hasHeightForWidth())
        self.page_footer.setSizePolicy(sizePolicy5)
        self.page_footer.setMinimumSize(QSize(0, 150))
        self.page_footer.setMaximumSize(QSize(16777215, 500))
        self.page_footer.setFrameShape(QFrame.Shape.NoFrame)
        self.page_footer.setFrameShadow(QFrame.Shadow.Raised)
        self.page_footer_layout = QVBoxLayout(self.page_footer)
        self.page_footer_layout.setSpacing(0)
        self.page_footer_layout.setObjectName(u"page_footer_layout")
        self.page_footer_layout.setSizeConstraint(QLayout.SizeConstraint.SetMinAndMaxSize)
        self.page_footer_layout.setContentsMargins(0, 0, 0, 1)
        self.console = QTextBrowser(self.page_footer)
        self.console.setObjectName(u"console")
        sizePolicy6 = QSizePolicy(QSizePolicy.Policy.Ignored, QSizePolicy.Policy.MinimumExpanding)
        sizePolicy6.setHorizontalStretch(0)
        sizePolicy6.setVerticalStretch(0)
        sizePolicy6.setHeightForWidth(self.console.sizePolicy().hasHeightForWidth())
        self.console.setSizePolicy(sizePolicy6)
        self.console.setMinimumSize(QSize(0, 150))
        self.console.setBaseSize(QSize(832, 200))
        font2 = QFont()
        font2.setFamilies([u"Consolas"])
        font2.setStyleStrategy(QFont.PreferAntialias)
        self.console.setFont(font2)
        self.console.setStyleSheet(u"")
        self.console.setFrameShape(QFrame.Shape.NoFrame)
        self.console.setFrameShadow(QFrame.Shadow.Raised)
        self.console.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOn)
        self.console.setSizeAdjustPolicy(QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)
        self.console.setAutoFormatting(QTextEdit.AutoFormattingFlag.AutoAll)
        self.console.setTabChangesFocus(True)
        self.console.setReadOnly(True)
        self.console.setHtml(u"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><meta charset=\"utf-8\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"hr { height: 1px; border-width: 0; }\n"
"li.unchecked::marker { content: \"\\2610\"; }\n"
"li.checked::marker { content: \"\\2612\"; }\n"
"</style></head><body style=\" font-family:'Consolas'; font-size:9pt; font-weight:400; font-style:normal;\">\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:'Cascadia Mono';\"><br /></p></body></html>")
        self.console.setOpenExternalLinks(True)

        self.page_footer_layout.addWidget(self.console)

        self.content_footer_splitter.addWidget(self.page_footer)

        self.page_frame_layout.addWidget(self.content_footer_splitter)


        self.center_frame_layout.addWidget(self.page_frame)

        self.right_menu_frame = QFrame(self.center_frame)
        self.right_menu_frame.setObjectName(u"right_menu_frame")
        sizePolicy7 = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.MinimumExpanding)
        sizePolicy7.setHorizontalStretch(0)
        sizePolicy7.setVerticalStretch(0)
        sizePolicy7.setHeightForWidth(self.right_menu_frame.sizePolicy().hasHeightForWidth())
        self.right_menu_frame.setSizePolicy(sizePolicy7)
        self.right_menu_frame.setMinimumSize(QSize(0, 0))
        self.right_menu_frame.setMaximumSize(QSize(0, 16777215))
        self.right_menu_frame.setBaseSize(QSize(0, 0))
        self.right_menu_frame.setFrameShape(QFrame.Shape.StyledPanel)
        self.right_menu_frame.setFrameShadow(QFrame.Shadow.Sunken)
        self.right_menu_frame.setLineWidth(1)
        self.right_menu_frame_layout = QVBoxLayout(self.right_menu_frame)
        self.right_menu_frame_layout.setObjectName(u"right_menu_frame_layout")
        self.right_menu_frame_layout.setContentsMargins(0, 0, 4, 5)
        self.right_menu_scroll_area = QScrollArea(self.right_menu_frame)
        self.right_menu_scroll_area.setObjectName(u"right_menu_scroll_area")
        sizePolicy8 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.MinimumExpanding)
        sizePolicy8.setHorizontalStretch(0)
        sizePolicy8.setVerticalStretch(0)
        sizePolicy8.setHeightForWidth(self.right_menu_scroll_area.sizePolicy().hasHeightForWidth())
        self.right_menu_scroll_area.setSizePolicy(sizePolicy8)
        self.right_menu_scroll_area.setMinimumSize(QSize(0, 0))
        self.right_menu_scroll_area.setFrameShape(QFrame.Shape.NoFrame)
        self.right_menu_scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.right_menu_scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.right_menu_scroll_area.setSizeAdjustPolicy(QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)
        self.right_menu_scroll_area.setWidgetResizable(True)
        self.right_menu_scroll_area_widgets = QWidget()
        self.right_menu_scroll_area_widgets.setObjectName(u"right_menu_scroll_area_widgets")
        self.right_menu_scroll_area_widgets.setGeometry(QRect(0, 0, 95, 411))
        self.right_menu_scroll_area_widgets.setMinimumSize(QSize(0, 0))
        self.right_menu_scroll_area_widgets_layout = QVBoxLayout(self.right_menu_scroll_area_widgets)
        self.right_menu_scroll_area_widgets_layout.setSpacing(2)
        self.right_menu_scroll_area_widgets_layout.setObjectName(u"right_menu_scroll_area_widgets_layout")
        self.right_menu_scroll_area_widgets_layout.setContentsMargins(2, 2, 0, 0)
        self.right_menu_content_frame = QFrame(self.right_menu_scroll_area_widgets)
        self.right_menu_content_frame.setObjectName(u"right_menu_content_frame")
        sizePolicy9 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.MinimumExpanding)
        sizePolicy9.setHorizontalStretch(0)
        sizePolicy9.setVerticalStretch(0)
        sizePolicy9.setHeightForWidth(self.right_menu_content_frame.sizePolicy().hasHeightForWidth())
        self.right_menu_content_frame.setSizePolicy(sizePolicy9)
        self.right_menu_content_frame.setFrameShadow(QFrame.Shadow.Raised)
        self.right_menu_content_frame_layout = QVBoxLayout(self.right_menu_content_frame)
        self.right_menu_content_frame_layout.setSpacing(0)
        self.right_menu_content_frame_layout.setObjectName(u"right_menu_content_frame_layout")
        self.right_menu_content_frame_layout.setContentsMargins(0, 0, 0, 0)
        self.right_menu_top_content_sw = QStackedWidget(self.right_menu_content_frame)
        self.right_menu_top_content_sw.setObjectName(u"right_menu_top_content_sw")
        sizePolicy9.setHeightForWidth(self.right_menu_top_content_sw.sizePolicy().hasHeightForWidth())
        self.right_menu_top_content_sw.setSizePolicy(sizePolicy9)
        self.right_menu_top_content_sw.setMinimumSize(QSize(0, 100))
        self.dummy_page_t = QWidget()
        self.dummy_page_t.setObjectName(u"dummy_page_t")
        sizePolicy9.setHeightForWidth(self.dummy_page_t.sizePolicy().hasHeightForWidth())
        self.dummy_page_t.setSizePolicy(sizePolicy9)
        self.dummy_layout2 = QVBoxLayout(self.dummy_page_t)
        self.dummy_layout2.setObjectName(u"dummy_layout2")
        self.pushButton_6 = QPushButton(self.dummy_page_t)
        self.pushButton_6.setObjectName(u"pushButton_6")

        self.dummy_layout2.addWidget(self.pushButton_6)

        self.verticalSpacer_2 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.dummy_layout2.addItem(self.verticalSpacer_2)

        self.right_menu_top_content_sw.addWidget(self.dummy_page_t)

        self.right_menu_content_frame_layout.addWidget(self.right_menu_top_content_sw)

        self.right_menu_bottom_content_sw = QStackedWidget(self.right_menu_content_frame)
        self.right_menu_bottom_content_sw.setObjectName(u"right_menu_bottom_content_sw")
        sizePolicy9.setHeightForWidth(self.right_menu_bottom_content_sw.sizePolicy().hasHeightForWidth())
        self.right_menu_bottom_content_sw.setSizePolicy(sizePolicy9)
        self.right_menu_bottom_content_sw.setMinimumSize(QSize(0, 100))
        self.dummy_page_b = QWidget()
        self.dummy_page_b.setObjectName(u"dummy_page_b")
        sizePolicy9.setHeightForWidth(self.dummy_page_b.sizePolicy().hasHeightForWidth())
        self.dummy_page_b.setSizePolicy(sizePolicy9)
        self.dummy_page_b.setStyleSheet(u"")
        self.dummy_layout = QVBoxLayout(self.dummy_page_b)
        self.dummy_layout.setObjectName(u"dummy_layout")
        self.pushButton = QPushButton(self.dummy_page_b)
        self.pushButton.setObjectName(u"pushButton")
        sizePolicy10 = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        sizePolicy10.setHorizontalStretch(0)
        sizePolicy10.setVerticalStretch(0)
        sizePolicy10.setHeightForWidth(self.pushButton.sizePolicy().hasHeightForWidth())
        self.pushButton.setSizePolicy(sizePolicy10)
        self.pushButton.setMinimumSize(QSize(0, 50))

        self.dummy_layout.addWidget(self.pushButton)

        self.verticalSpacer = QSpacerItem(20, 5, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.MinimumExpanding)

        self.dummy_layout.addItem(self.verticalSpacer)

        self.right_menu_bottom_content_sw.addWidget(self.dummy_page_b)

        self.right_menu_content_frame_layout.addWidget(self.right_menu_bottom_content_sw)


        self.right_menu_scroll_area_widgets_layout.addWidget(self.right_menu_content_frame)

        self.right_menu_scroll_area.setWidget(self.right_menu_scroll_area_widgets)

        self.right_menu_frame_layout.addWidget(self.right_menu_scroll_area)


        self.center_frame_layout.addWidget(self.right_menu_frame)


        self.main_frame_layout.addWidget(self.center_frame, 1, 1, 1, 1)

        self.left_menu_frame = QFrame(self.main_frame)
        self.left_menu_frame.setObjectName(u"left_menu_frame")
        self.left_menu_frame.setStyleSheet(u"")
        self.left_menu_frame.setFrameShape(QFrame.Shape.NoFrame)
        self.left_menu_frame.setFrameShadow(QFrame.Shadow.Raised)
        self.left_menu_frame_layout = QVBoxLayout(self.left_menu_frame)
        self.left_menu_frame_layout.setSpacing(0)
        self.left_menu_frame_layout.setObjectName(u"left_menu_frame_layout")
        self.left_menu_frame_layout.setSizeConstraint(QLayout.SizeConstraint.SetDefaultConstraint)
        self.left_menu_frame_layout.setContentsMargins(3, 0, 0, 0)
        self.left_menu_top_subframe = QFrame(self.left_menu_frame)
        self.left_menu_top_subframe.setObjectName(u"left_menu_top_subframe")
        self.left_menu_top_subframe.setFrameShape(QFrame.Shape.NoFrame)
        self.left_menu_top_subframe.setFrameShadow(QFrame.Shadow.Raised)
        self.left_menu_top_subframe_layout = QVBoxLayout(self.left_menu_top_subframe)
        self.left_menu_top_subframe_layout.setSpacing(15)
        self.left_menu_top_subframe_layout.setObjectName(u"left_menu_top_subframe_layout")
        self.left_menu_top_subframe_layout.setContentsMargins(3, 5, 3, 0)
        self.frame = QFrame(self.left_menu_top_subframe)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.Shape.NoFrame)
        self.frame.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout = QHBoxLayout(self.frame)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(4, 0, 0, 0)
        self.title_icon_label = QLabel(self.frame)
        self.title_icon_label.setObjectName(u"title_icon_label")
        sizePolicy11 = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        sizePolicy11.setHorizontalStretch(0)
        sizePolicy11.setVerticalStretch(0)
        sizePolicy11.setHeightForWidth(self.title_icon_label.sizePolicy().hasHeightForWidth())
        self.title_icon_label.setSizePolicy(sizePolicy11)
        self.title_icon_label.setMinimumSize(QSize(25, 25))
        self.title_icon_label.setMaximumSize(QSize(25, 25))
        self.title_icon_label.setPixmap(QPixmap(u"../../assets/icons/icon.ico"))
        self.title_icon_label.setScaledContents(True)
        self.title_icon_label.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.horizontalLayout.addWidget(self.title_icon_label)

        self.title_label = QLabel(self.frame)
        self.title_label.setObjectName(u"title_label")
        self.title_label.setMinimumSize(QSize(0, 30))
        font3 = QFont()
        font3.setBold(True)
        font3.setItalic(False)
        font3.setStyleStrategy(QFont.PreferAntialias)
        self.title_label.setFont(font3)
        self.title_label.setStyleSheet(u"font: bold")
        self.title_label.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignVCenter)
        self.title_label.setMargin(0)
        self.title_label.setIndent(5)

        self.horizontalLayout.addWidget(self.title_label)


        self.left_menu_top_subframe_layout.addWidget(self.frame)

        self.toggle_left_menu_button = QPushButton(self.left_menu_top_subframe)
        self.toggle_left_menu_button.setObjectName(u"toggle_left_menu_button")
        sizePolicy11.setHeightForWidth(self.toggle_left_menu_button.sizePolicy().hasHeightForWidth())
        self.toggle_left_menu_button.setSizePolicy(sizePolicy11)
        self.toggle_left_menu_button.setMinimumSize(QSize(64, 45))
        self.toggle_left_menu_button.setMaximumSize(QSize(64, 45))
        icon1 = QIcon()
        icon1.addFile(u"../../../../../../Users/peter/.designer/assets/icons/menu_stripes.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.toggle_left_menu_button.setIcon(icon1)
        self.toggle_left_menu_button.setIconSize(QSize(32, 32))
        self.toggle_left_menu_button.setCheckable(False)
        self.toggle_left_menu_button.setFlat(True)

        self.left_menu_top_subframe_layout.addWidget(self.toggle_left_menu_button)


        self.left_menu_frame_layout.addWidget(self.left_menu_top_subframe)

        self.left_menu_middle_subframe = QFrame(self.left_menu_frame)
        self.left_menu_middle_subframe.setObjectName(u"left_menu_middle_subframe")
        self.left_menu_middle_subframe.setMinimumSize(QSize(0, 0))
        self.left_menu_middle_subframe.setFrameShape(QFrame.Shape.NoFrame)
        self.left_menu_middle_subframe.setFrameShadow(QFrame.Shadow.Raised)
        self.left_menu_middle_subframe_layout = QVBoxLayout(self.left_menu_middle_subframe)
        self.left_menu_middle_subframe_layout.setObjectName(u"left_menu_middle_subframe_layout")
        self.left_menu_middle_subframe_layout.setContentsMargins(3, -1, 3, 0)

        self.left_menu_frame_layout.addWidget(self.left_menu_middle_subframe)

        self.left_menu_bottom_subframe = QFrame(self.left_menu_frame)
        self.left_menu_bottom_subframe.setObjectName(u"left_menu_bottom_subframe")
        self.left_menu_bottom_subframe.setFrameShape(QFrame.Shape.NoFrame)
        self.left_menu_bottom_subframe.setFrameShadow(QFrame.Shadow.Raised)
        self.left_menu_bottom_subframe_layout = QVBoxLayout(self.left_menu_bottom_subframe)
        self.left_menu_bottom_subframe_layout.setSpacing(5)
        self.left_menu_bottom_subframe_layout.setObjectName(u"left_menu_bottom_subframe_layout")
        self.left_menu_bottom_subframe_layout.setContentsMargins(3, 0, 3, 10)
        self.left_menu_bottom_vspacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.left_menu_bottom_subframe_layout.addItem(self.left_menu_bottom_vspacer)

        self.settings_button = QPushButton(self.left_menu_bottom_subframe)
        self.settings_button.setObjectName(u"settings_button")
        self.settings_button.setMinimumSize(QSize(64, 50))
        self.settings_button.setMaximumSize(QSize(64, 16777215))
        icon2 = QIcon()
        icon2.addFile(u"../../../../../../Users/peter/.designer/assets/icons/settings.png", QSize(), QIcon.Mode.Normal, QIcon.State.On)
        self.settings_button.setIcon(icon2)
        self.settings_button.setIconSize(QSize(32, 32))
        self.settings_button.setCheckable(True)
        self.settings_button.setChecked(False)
        self.settings_button.setFlat(True)

        self.left_menu_bottom_subframe_layout.addWidget(self.settings_button)


        self.left_menu_frame_layout.addWidget(self.left_menu_bottom_subframe)


        self.main_frame_layout.addWidget(self.left_menu_frame, 0, 0, 2, 1)

        self.main_frame_layout.setRowStretch(0, 1)
        self.main_frame_layout.setRowStretch(1, 10)
        self.main_frame_layout.setColumnStretch(0, 1)
        self.main_frame_layout.setColumnStretch(1, 20)

        self.central_widget_layout.addWidget(self.main_frame)

        MainWindow.setCentralWidget(self.central_widget)
        QWidget.setTabOrder(self.toggle_left_menu_button, self.settings_button)
        QWidget.setTabOrder(self.settings_button, self.listView_2)
        QWidget.setTabOrder(self.listView_2, self.search_bar_line_edit)
        QWidget.setTabOrder(self.search_bar_line_edit, self.minimize_button)
        QWidget.setTabOrder(self.minimize_button, self.restore_max_button)
        QWidget.setTabOrder(self.restore_max_button, self.close_button)
        QWidget.setTabOrder(self.close_button, self.right_menu_scroll_area)
        QWidget.setTabOrder(self.right_menu_scroll_area, self.pushButton_6)
        QWidget.setTabOrder(self.pushButton_6, self.pushButton)
        QWidget.setTabOrder(self.pushButton, self.console)

        self.retranslateUi(MainWindow)

        self.page_stacked_widget.setCurrentIndex(0)
        self.toggle_left_menu_button.setDefault(False)


        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"Conan App Launcher", None))
        self.page_title.setText(QCoreApplication.translate("MainWindow", u"Page Title", None))
        self.close_button.setText("")
        self.restore_max_button.setText("")
        self.minimize_button.setText("")
        self.search_bar_line_edit.setInputMask("")
        self.search_bar_line_edit.setText("")
        self.search_bar_line_edit.setPlaceholderText("")
        self.pushButton_6.setText(QCoreApplication.translate("MainWindow", u"PushButton", None))
        self.pushButton.setText(QCoreApplication.translate("MainWindow", u"PushButton", None))
        self.title_icon_label.setText("")
        self.title_label.setText(QCoreApplication.translate("MainWindow", u"App Title", None))
#if QT_CONFIG(tooltip)
        self.toggle_left_menu_button.setToolTip(QCoreApplication.translate("MainWindow", u"Expand/Collapse Menu", None))
#endif // QT_CONFIG(tooltip)
        self.toggle_left_menu_button.setText("")
#if QT_CONFIG(tooltip)
        self.settings_button.setToolTip(QCoreApplication.translate("MainWindow", u"Show Settings Menu", None))
#endif // QT_CONFIG(tooltip)
        self.settings_button.setText("")
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'reorder_dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QDialog, QDialogButtonBox,
    QFrame, QGridLayout, QLayout, QListView,
    QPushButton, QSizePolicy, QSpacerItem, QVBoxLayout,
    QWidget)

class Ui_rearrange_dialog(object):
    def setupUi(self, rearrange_dialog):
        if not rearrange_dialog.objectName():
            rearrange_dialog.setObjectName(u"rearrange_dialog")
        rearrange_dialog.resize(443, 331)
        self.gridLayout = QGridLayout(rearrange_dialog)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setSizeConstraint(QLayout.SetMinAndMaxSize)
        self.list_view = QListView(rearrange_dialog)
        self.list_view.setObjectName(u"list_view")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.list_view.sizePolicy().hasHeightForWidth())
        self.list_view.setSizePolicy(sizePolicy)

        self.gridLayout.addWidget(self.list_view, 0, 0, 1, 1)

        self.vertical_layout = QVBoxLayout()
        self.vertical_layout.setObjectName(u"vertical_layout")
        self.move_up_button = QPushButton(rearrange_dialog)
        self.move_up_button.setObjectName(u"move_up_button")

        self.vertical_layout.addWidget(self.move_up_button)

        self.line = QFrame(rearrange_dialog)
        self.line.setObjectName(u"line")
        self.line.setFrameShape(QFrame.Shape.HLine)
        self.line.setFrameShadow(QFrame.Shadow.Sunken)

        self.vertical_layout.addWidget(self.line)

        self.move_down_button = QPushButton(rearrange_dialog)
        self.move_down_button.setObjectName(u"move_down_button")

        self.vertical_layout.addWidget(self.move_down_button)

        self.vertical_spacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.vertical_layout.addItem(self.vertical_spacer)


        self.gridLayout.addLayout(self.vertical_layout, 0, 1, 1, 1)

        self.button_box = QDialogButtonBox(rearrange_dialog)
        self.button_box.setObjectName(u"button_box")
        self.button_box.setOrientation(Qt.Horizontal)
        self.button_box.setStandardButtons(QDialogButtonBox.Ok)
        self.button_box.setCenterButtons(False)

        self.gridLayout.addWidget(self.button_box, 1, 0, 1, 1)


        self.retranslateUi(rearrange_dialog)
        self.button_box.accepted.connect(rearrange_dialog.accept)
        self.button_box.rejected.connect(rearrange_dialog.reject)

        QMetaObject.connectSlotsByName(rearrange_dialog)
    # setupUi

    def retranslateUi(self, rearrange_dialog):
        rearrange_dialog.setWindowTitle(QCoreApplication.translate("rearrange_dialog", u"Rearrange App Grid", None))
        self.move_up_button.setText(QCoreApplication.translate("rearrange_dialog", u"Move up", None))
        self.move_down_button.setText(QCoreApplication.translate("rearrange_dialog", u"Move down", None))
    # retranslateUi




import logging
from threading import Lock

from PySide6.QtCore import SignalInstance

from conan_explorer import CONAN_LOG_PREFIX
from conan_explorer.app.logger import Logger


class QtLogHandler(logging.Handler):
    """
    This log handler sends a logger string to a qt widget.
    update_signal needs str as argument.
    """

    _lock = Lock()
    _formatter = logging.Formatter(r"%(levelname)s: %(message)s")

    def __init__(self, update_signal: SignalInstance):
        super().__init__(logging.DEBUG)
        self._update_signal = update_signal

    def emit(self, record: logging.LogRecord):
        # don't access the qt object directly, since updates will only work
        # correctly in main loop, so instead send a PyQt Signal with the text to the Ui

        if not record.message.startswith(CONAN_LOG_PREFIX):
            message = self.format(record)
        else:
            # remove log formatting for conan logs
            message = record.message.replace(CONAN_LOG_PREFIX, "")
        if message and self._lock:  # one log at a time
            with self._lock:
                try:
                    self._update_signal.emit(message)
                except Exception:
                    print("QT Logger errored")  # don't log here with logger...


def init_qt_logger(logger: Logger, name: str, update_signal: SignalInstance):
    """Redirects the logger to QT widget.
    Needs to be called when GUI objects are available.
    update_signal needs str as argument.
    """
    qt_handler = QtLogHandler(update_signal)
    qt_handler.set_name(name)
    log_debug_level = logging.INFO
    qt_handler.setLevel(log_debug_level)
    qt_handler.setFormatter(qt_handler._formatter)

    logger.addHandler(qt_handler)


def remove_qt_logger(logger: Logger, name: str) -> bool:
    """Remove qt logger (to be called before gui closes)"""
    for handler in logger.handlers:
        if handler.get_name() == name:
            logger.removeHandler(handler)
            if handler:
                del handler
            return True
    return False



import importlib
import sys
from pathlib import Path
from typing import List, Optional

from packaging import specifiers
from PySide6.QtCore import QObject, Signal, SignalInstance
from PySide6.QtWidgets import QWidget

import conan_explorer.app as app
from conan_explorer import BUILT_IN_PLUGIN, Version, base_path, conan_version
from conan_explorer.app.logger import Logger
from conan_explorer.settings import PLUGINS_SECTION_NAME

from .file import PluginFile
from .types import PluginDescription, PluginInterfaceV1


class PluginHandler(QObject):
    # Both signals need to be connected in the main gui
    load_plugin: SignalInstance = Signal(PluginDescription)  # type: ignore
    unload_plugin: SignalInstance = Signal(str)  # type: ignore

    def __init__(self, parent: QWidget, base_signals, page_widgets) -> None:
        super().__init__(parent)
        self._base_signals = base_signals
        self._page_widgets = page_widgets
        self._active_plugins: List[PluginInterfaceV1] = []

    def load_all_plugins(self):
        # load built-in from dynamic path
        for plugin_group_name in app.active_settings.get_settings_from_node(
            PLUGINS_SECTION_NAME
        ):
            plugin_path = app.active_settings.get_string(plugin_group_name)
            if plugin_group_name == BUILT_IN_PLUGIN:
                # fix potentially outdated setting
                correct_plugin_path = str(base_path / "ui" / "plugins.ini")
                if plugin_path != correct_plugin_path:
                    app.active_settings.add(
                        BUILT_IN_PLUGIN, correct_plugin_path, PLUGINS_SECTION_NAME
                    )
                    plugin_path = correct_plugin_path
            self._load_plugins_from_file(plugin_path)

    def reload_all_plugins(self):
        for plugin in self._active_plugins:
            self.reload_plugin(plugin.plugin_description.name)

    def get_plugin_descr_from_name(self, plugin_name: str) -> Optional[PluginDescription]:
        plugins: List[PluginDescription] = self.get_same_file_plugins_from_name(plugin_name)
        for plugin in plugins:
            if plugin.name == plugin_name:
                return plugin

    def get_same_file_plugins_from_name(self, plugin_name: str) -> List[PluginDescription]:
        for plugin_group_name in app.active_settings.get_settings_from_node(
            PLUGINS_SECTION_NAME
        ):
            plugin_path = app.active_settings.get_string(plugin_group_name)
            file_plugins = PluginFile.read(plugin_path)
            for plugin in file_plugins:
                if plugin.name == plugin_name:
                    return file_plugins
        return []

    def add_plugin(self, plugin_path: str):
        PluginFile.register(plugin_path)
        plugins = self._load_plugins_from_file(plugin_path)
        for plugin in plugins:
            self.load_plugin.emit(plugin)
            plugin.load_signal.emit()

    def remove_plugin(self, plugin_path: str):
        PluginFile.unregister(plugin_path)
        file_plugins = PluginFile.read(plugin_path)
        for plugin in file_plugins:
            self._unload_plugin(plugin)

    def reload_plugin(self, plugin_name: str):
        plugin_descr = self.get_plugin_descr_from_name(plugin_name)
        assert plugin_descr
        self._unload_plugin(plugin_descr)
        plugin = self._load_plugin(plugin_descr, reload=True)
        if not plugin:
            return
        self.load_plugin.emit(plugin)
        plugin.load_signal.emit()

    def _load_plugins_from_file(self, plugin_path: str) -> List[PluginInterfaceV1]:
        file_plugin_descrs = PluginFile.read(plugin_path)
        loaded_plugins = []
        for plugin_descr in file_plugin_descrs:
            loaded_plugins.append(self._load_plugin(plugin_descr))
        return loaded_plugins

    def _load_plugin(
        self, plugin: PluginDescription, reload=False
    ) -> Optional[PluginInterfaceV1]:
        loaded_plugin = None
        if self.is_plugin_enabled(plugin):
            try:
                import_path = Path(plugin.import_path)
                sys.path.append(str(import_path.parent))
                module_ = importlib.import_module(import_path.stem)
                if reload:
                    # get all modules to re-import
                    modules_to_del = [
                        module for module in sys.modules if import_path.stem in module
                    ]
                    # importlib reload does not work with relative imports
                    for module_to_del in modules_to_del:
                        del sys.modules[module_to_del]
                    module_ = importlib.import_module(import_path.stem)

                class_ = getattr(module_, plugin.plugin_class)
                plugin_object: PluginInterfaceV1 = class_(
                    self.parent(), plugin, self._base_signals, self._page_widgets
                )
                loaded_plugin = plugin_object
            except Exception as e:
                Logger().error(f"Can't load plugin {plugin.name}: {str(e)}")
        else:
            Logger().info(
                f"Can't load plugin {plugin.name}."
                f"Conan version restriction {plugin.conan_versions} applies."
            )
        if loaded_plugin:
            self._active_plugins.append(loaded_plugin)
        return loaded_plugin

    def post_load_plugins(self):
        for plugin in self._active_plugins:
            self.load_plugin.emit(plugin)
            plugin.load_signal.emit()

    def _unload_plugin(self, plugin: PluginDescription):
        plugin_widget = self.get_plugin_by_description(plugin)
        if not plugin_widget:
            Logger().error(f"Cannot get plugin {plugin.name} for unload")
            return
        self._active_plugins.remove(plugin_widget)
        plugin_widget.close()
        plugin_widget.deleteLater()
        self.unload_plugin.emit(plugin.name)

    def get_plugin_by_description(
        self, plugin_descr: PluginDescription
    ) -> Optional[PluginInterfaceV1]:
        """Return Plugin Object by full plugin description"""
        for plugin in self._active_plugins:
            if isinstance(plugin, PluginInterfaceV1):
                try:
                    if plugin.plugin_description == plugin_descr:
                        return plugin
                except Exception as e:
                    Logger().error(f"Can't retrive plugin information: {str(e)}")

    @staticmethod
    def eval_conan_version_spec(spec: str, conan_version: Version = conan_version) -> bool:
        if not spec:
            return True
        specs = specifiers.Specifier(spec)
        return specs.contains(conan_version)

    @staticmethod
    def is_plugin_enabled(plugin: PluginDescription):
        result = PluginHandler.eval_conan_version_spec(plugin.conan_versions)
        return result



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'about.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractScrollArea, QApplication, QFrame, QLabel,
    QScrollArea, QSizePolicy, QVBoxLayout, QWidget)

class Ui_Form(object):
    def setupUi(self, Form):
        if not Form.objectName():
            Form.setObjectName(u"Form")
        Form.resize(663, 456)
        Form.setAutoFillBackground(True)
        Form.setStyleSheet(u"")
        self.verticalLayout_2 = QVBoxLayout(Form)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.about_scroll_area = QScrollArea(Form)
        self.about_scroll_area.setObjectName(u"about_scroll_area")
        self.about_scroll_area.setFrameShape(QFrame.NoFrame)
        self.about_scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.about_scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.about_scroll_area.setSizeAdjustPolicy(QAbstractScrollArea.AdjustToContentsOnFirstShow)
        self.about_scroll_area.setWidgetResizable(True)
        self.about_contents = QWidget()
        self.about_contents.setObjectName(u"about_contents")
        self.about_contents.setGeometry(QRect(0, 0, 646, 456))
        self.verticalLayout = QVBoxLayout(self.about_contents)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.logo_label = QLabel(self.about_contents)
        self.logo_label.setObjectName(u"logo_label")
        self.logo_label.setMinimumSize(QSize(100, 100))
        self.logo_label.setMaximumSize(QSize(100, 100))
        self.logo_label.setPixmap(QPixmap(u"../../../assets/icons/icon.ico"))
        self.logo_label.setScaledContents(True)
        self.logo_label.setWordWrap(False)

        self.verticalLayout.addWidget(self.logo_label)

        self.about_label = QLabel(self.about_contents)
        self.about_label.setObjectName(u"about_label")
        self.about_label.setTextFormat(Qt.RichText)
        self.about_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.about_label.setWordWrap(False)
        self.about_label.setOpenExternalLinks(True)

        self.verticalLayout.addWidget(self.about_label)

        self.verticalLayout.setStretch(0, 1)
        self.verticalLayout.setStretch(1, 1)
        self.about_scroll_area.setWidget(self.about_contents)

        self.verticalLayout_2.addWidget(self.about_scroll_area)


        self.retranslateUi(Form)

        QMetaObject.connectSlotsByName(Form)
    # setupUi

    def retranslateUi(self, Form):
        Form.setWindowTitle(QCoreApplication.translate("Form", u"Form", None))
        self.logo_label.setText("")
        self.about_label.setText(QCoreApplication.translate("Form", u"<html><head/><body><p><br/>Hallo </p></body></html>", None))
    # retranslateUi




from typing import Dict, List, Optional

from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import Qt, SignalInstance, Slot
from typing_extensions import override

import conan_explorer.app as app
from conan_explorer.app import LoaderGui  # using global module pattern
from conan_explorer.app.logger import Logger
from conan_explorer.conan_wrapper import ConanApiFactory
from conan_explorer.conan_wrapper.types import ConanPkg, ConanRef
from conan_explorer.ui.common import (
    TreeModel,
    TreeModelItem,
    get_platform_icon,
    get_themed_asset_icon,
)

REF_TYPE = 0
PROFILE_TYPE = 1


class SearchedPackageTreeItem(TreeModelItem):
    """
    Represents a tree item of a Conan pkg.
    1. ref/id 2. remote 3. quick profile
    """

    def __init__(
        self,
        data: List[str],
        parent=None,
        pkg_info: Optional[ConanPkg] = None,
        item_type=REF_TYPE,
        lazy_loading=False,
        installed=False,
        empty=False,
    ):
        super().__init__(data, parent, lazy_loading=lazy_loading)
        self.type = item_type
        self.pkg_info = pkg_info
        self.is_installed = installed
        self.empty = empty  # indicates a "no result" item, which must be handled separately
        self.child_items: List[SearchedPackageTreeItem] = []

    # @override
    def load_children(self, update_sig):
        # can't call super method: fetching would finish early
        self.child_items = []

        pkgs_to_be_added: Dict[str, SearchedPackageTreeItem] = {}
        for remote in self.data(1).split(","):
            update_sig.emit("Loading from " + str(remote))
            recipe_ref = self.data(0)
            # cross reference with installed packages
            infos = app.conan_api.get_local_pkgs_from_ref(ConanRef.loads(recipe_ref))
            installed_ids = [info.get("id") for info in infos]
            packages = app.conan_api.get_remote_pkgs_from_ref(
                ConanRef.loads(recipe_ref), remote
            )
            for pkg in packages:
                pkg_id = pkg.get("id", "")
                if pkg_id in pkgs_to_be_added.keys():  # package already found in another remote
                    pkgs_to_be_added[pkg_id].item_data[1] += "," + remote
                    continue
                installed = False
                if pkg_id in installed_ids:
                    installed = True
                pkgs_to_be_added[pkg_id] = SearchedPackageTreeItem(
                    [
                        pkg_id,
                        remote,
                        ConanApiFactory().build_conan_profile_name_alias(
                            pkg.get("settings", {})
                        ),
                    ],
                    self,
                    pkg,
                    PROFILE_TYPE,
                    False,
                    installed,
                )
        for pkg in pkgs_to_be_added.values():
            self.child_items.append(pkg)
        if not self.child_items:
            self.child_items.append(
                SearchedPackageTreeItem(
                    ["No package found", "", ""], self, {}, PROFILE_TYPE, empty=True
                )
            )
        self.is_loaded = True

    @override
    def child_count(self) -> int:
        if self.type == REF_TYPE:
            return len(self.child_items) if len(self.child_items) > 0 else 1
        elif self.type == PROFILE_TYPE:
            return 0  # no child
        return 0  # for safety

    def get_conan_ref(self) -> str:
        if self.type == REF_TYPE:
            return self.data(0)
        elif self.type == PROFILE_TYPE:
            parent = self.parent()
            if parent is None:
                return ""
            return parent.data(0) + ":" + self.data(0)
        return ""


class PkgSearchModel(TreeModel):
    def __init__(
        self,
        conan_pkg_installed: Optional[SignalInstance] = None,
        conan_pkg_removed: Optional[SignalInstance] = None,
        *args,
        **kwargs,
    ):
        super(PkgSearchModel, self).__init__(*args, **kwargs)
        self.root_item = SearchedPackageTreeItem(["Packages", "Remote(s)", "Quick Profile"])
        self.proxy_model = QtCore.QSortFilterProxyModel()  # for sorting
        self.proxy_model.setDynamicSortFilter(True)
        self.proxy_model.setSourceModel(self)
        self._loader_widget_parent = None
        if conan_pkg_installed:
            conan_pkg_installed.connect(self.mark_pkg_as_installed)
        if conan_pkg_removed:
            conan_pkg_removed.connect(self.mark_pkg_as_not_installed)

    def setup_model_data(self, search_query: str, remotes: List[str], update_sig):
        # needs to be ConanRef, so we can check with get_all_local_refs directly
        self.clear_items()
        self.beginResetModel()
        recipes_with_remotes: Dict[ConanRef, str] = {}
        for remote in remotes:
            update_sig.emit("Searching in " + str(remote))
            recipe_list = app.conan_api.search_recipes_in_remotes(
                f"{search_query}*", remote_name=remote
            )
            for recipe in recipe_list:
                current_value = recipes_with_remotes.get(recipe, "")
                if current_value:
                    recipes_with_remotes[recipe] = current_value + "," + remote
                else:  # element 0
                    recipes_with_remotes[recipe] = remote

        if not recipes_with_remotes:
            self.root_item.append_child(
                SearchedPackageTreeItem(
                    ["No package found!", "", ""],
                    self.root_item,
                    None,
                    PROFILE_TYPE,
                    empty=True,
                )
            )
            return

        # add info if it is installed
        installed_refs = app.conan_api.get_all_local_refs()
        for recipe in recipes_with_remotes:
            recipe_remotes = recipes_with_remotes.get(recipe, "")
            installed = False
            if recipe in installed_refs:
                installed = True
            conan_item = SearchedPackageTreeItem(
                [str(recipe), recipe_remotes, ""],
                self.root_item,
                None,
                REF_TYPE,
                lazy_loading=True,
                installed=installed,
            )
            self.root_item.append_child(conan_item)
        self.endResetModel()

    @override
    def data(self, index: "QtCore.QModelIndex | QtCore.QPersistentModelIndex", role: int = 0):
        if not index.isValid():
            return None
        item: SearchedPackageTreeItem = index.internalPointer()  # type: ignore
        if role == Qt.ItemDataRole.DecorationRole:
            if index.column() != 0:  # only display icon for first column
                return
            if item.type == REF_TYPE:
                return QtGui.QIcon(get_themed_asset_icon("icons/package.svg"))
            if item.type == PROFILE_TYPE:
                profile_name = item.data(2)
                return get_platform_icon(profile_name)
        elif role == Qt.ItemDataRole.DisplayRole:
            return item.data(index.column())
        elif role == Qt.ItemDataRole.FontRole and item.is_installed:
            font = QtGui.QFont()
            font.setBold(True)
            return font
        return None

    def get_item_from_ref(self, conan_ref: str) -> Optional[SearchedPackageTreeItem]:
        for item in self.root_item.child_items:
            if item.item_data[0] == conan_ref:
                return item
        return None

    @Slot(str, str)
    def mark_pkg_as_installed(self, conan_ref: str, pkg_id: str):
        self._set_pkg_install_status(conan_ref, pkg_id, True)

    @Slot(str, str)
    def mark_pkg_as_not_installed(self, conan_ref: str, pkg_id: str):
        self._set_pkg_install_status(conan_ref, pkg_id, False)

    def _set_pkg_install_status(self, conan_ref: str, pkg_id: str, installed: bool):
        item = self.get_item_from_ref(conan_ref)
        if not item:
            return
        item.is_installed = installed
        pkg_items = item.child_items
        for pkg_item in pkg_items:
            if not pkg_item.pkg_info:
                return
            if installed and pkg_item.pkg_info.get("id", "") == pkg_id:
                Logger().debug(f"Set {pkg_id} as install status to {installed}")
                pkg_item.is_installed = installed
                break
            if not pkg_id and not installed:  # if ref was removed, all pkgs are deleted too
                pkg_item.is_installed = installed

    @override
    def fetchMore(self, index):
        item = index.internalPointer()
        loader = LoaderGui(self)
        self._loader_widget_parent = QtWidgets.QWidget()
        loader.load_for_blocking(
            self._loader_widget_parent,
            item.load_children,
            (loader.loading_string_signal,),
            loading_text="Loading Packages...",
        )
        loader.wait_for_finished()



"""
Contains all basic constants used in the application.
No imports from own modules allowed! This is done to resolve circular dependencies.
"""

import os
import platform
import shutil
from importlib.metadata import PackageNotFoundError, distribution
from pathlib import Path
from typing import TypeVar

from packaging.version import Version

PathLike = TypeVar("PathLike", str, Path)

### Global constants ###

# load metadata from package info - needs to be installed as editable in dev mode!
PKG_NAME = "conan_explorer"
APP_NAME = "Conan Explorer"

try:
    pkg_info = distribution(PKG_NAME)
    __version__ = pkg_info.version
    REPO_URL = pkg_info.metadata.get("home-page", "")  # type: ignore
    AUTHOR = pkg_info.metadata.get("author", "")  # type: ignore
except PackageNotFoundError:  # pragma: no cover
    # For local usecases, when there is no distribution
    __version__ = "1.0.0"
    REPO_URL = ""
    AUTHOR = ""

ICON_SIZE = 64  # Icon size (width and height) in pixels on an Applink
MAX_FONT_SIZE = 16
MIN_FONT_SIZE = 10
INVALID_PATH = "Unknown"
# used to indicate a conan reference is invalid
INVALID_CONAN_REF = "Invalid/0.0.1@NA/NA"
SETTINGS_FILE_NAME = "settings"  # for storing application settings
# for backwards compatibility till 1.4.1
LEGACY_SETTINGS_FILE_NAME = ".cal_config"
DEFAULT_UI_CFG_FILE_NAME = "ui_cfg"  # no extension from 1.4.1
# for backwards compatibility till 1.4.1
LEGACY_UI_CFG_FILE_NAME = "cal_ui.json"
CONAN_LOG_PREFIX = "CONAN: "  # logger uses this to indicate a log comes from Conan
BUILT_IN_PLUGIN = "built-in"

# Feature flags
# get versions directly from the custom cache
SEARCH_APP_VERSIONS_IN_LOCAL_CACHE = True
# get pkg paths directly from the custom cache TODO can't handle options
USE_LOCAL_CACHE_FOR_LOCAL_PKG_PATH = True
# use conan worker to also search for the package path
# works in a addition to USE_LOCAL_CACHE_FOR_LOCAL_PKG_PATH and also installs
USE_CONAN_WORKER_FOR_LOCAL_PKG_PATH_AND_INSTALL = True
AUTOCLOSE_SIDE_MENU = False
ENABLE_GUI_STYLES = False

# From ennvar DEBUG - 0: No debug, 1 = debug logging on, 2 = disable true async loading
DEBUG_LEVEL = int(os.getenv("CAL_DEBUG_LEVEL", "0"))

# Conan version
conan_pkg_info = distribution("conan")

conan_version = Version(conan_pkg_info.version)

# Paths to find folders - points to the folder of this file
# must be initialized later, otherwise setup.py can't parse this file

base_path = Path(__file__).absolute().parent
asset_path = base_path / "assets"
# to be used for all default paths of configuration files, which will be used
# for multiple versions (noninvasive storage, 1.X legacy will be moved)
# TODO: Still use conan_app_launcher - move to conan_explorer later
# and move to separate function
legacy_user_save_path = Path().home()
user_save_path = (
    Path(os.getenv("XDG_CONFIG_HOME", str(legacy_user_save_path / ".config")))
    / "conan_app_launcher"
    if platform.system() == "Linux"
    else Path(os.getenv("APPDATA", str(legacy_user_save_path))) / "conan_app_launcher"
)

# user path migration - move settings and default gui file
# ui file loading will handle patching the settings, if the default gui file was used
if user_save_path != legacy_user_save_path:
    os.makedirs(str(user_save_path), exist_ok=True)
    # don't copy if the migrated files already exist
    if (legacy_user_save_path / LEGACY_SETTINGS_FILE_NAME).exists() and not (
        user_save_path / (SETTINGS_FILE_NAME + ".ini")
    ).exists():
        print(
            (
                "INFO: Moving application settings file from"
                f"{str(user_save_path)} to {str(legacy_user_save_path)}"
            )
        )
        shutil.move(
            str(legacy_user_save_path / LEGACY_SETTINGS_FILE_NAME),
            str(user_save_path / (SETTINGS_FILE_NAME + ".ini")),
        )
    if (legacy_user_save_path / LEGACY_UI_CFG_FILE_NAME).exists() and not (
        user_save_path / (DEFAULT_UI_CFG_FILE_NAME + ".json")
    ).exists():
        print(
            "INFO: Moving default ui config file from "
            f"{str(user_save_path)} to {str(legacy_user_save_path)}"
        )
        shutil.move(
            str(legacy_user_save_path / LEGACY_UI_CFG_FILE_NAME),
            str(user_save_path / (DEFAULT_UI_CFG_FILE_NAME + ".json")),
        )



from pathlib import Path

from PySide6.QtWidgets import QDialog, QFileDialog

import conan_explorer.app as app
from conan_explorer.app.logger import Logger
from conan_explorer.settings import FILE_EDITOR_EXECUTABLE


class FileEditorSelDialog(QDialog):
    def __init__(self, parent) -> None:
        super().__init__(parent=parent)
        from .file_editor_selector_ui import Ui_Form

        self._ui = Ui_Form()
        self._ui.setupUi(self)
        self.setWindowTitle("File Editor Selection")

        self._ui.file_edit.setText(app.active_settings.get_string(FILE_EDITOR_EXECUTABLE))
        self._ui.browse_button.clicked.connect(self.on_browse_clicked)
        self._ui.button_box.accepted.connect(self.on_save)
        self._ui.button_box.rejected.connect(self.close)

    def on_browse_clicked(self):
        current_file = Path(self._ui.file_edit.text())
        path_to_open = Path().home
        if current_file.exists():
            path_to_open = current_file.parent
        dialog = QFileDialog(
            parent=self,
            caption="Select file editor executable",
            directory=str(path_to_open),
            filter="Executables (*)",
        )
        dialog.setFileMode(QFileDialog.FileMode.ExistingFile)
        if dialog.exec() == QFileDialog.DialogCode.Accepted:
            selected_path = Path(dialog.selectedFiles()[0])
            try:
                self._ui.file_edit.setText(str(selected_path))
            except Exception as e:
                # errors, if it does not resolve
                Logger().debug("Can't set selected path: %s.", str(e))

    def on_save(self):
        app.active_settings.set(FILE_EDITOR_EXECUTABLE, self._ui.file_edit.text())
        self.close()
